import unittest

from mock import Mock, patch

from twilio.rest.resources import Applications, Application


class ApplicationsTest(unittest.TestCase):
    def setUp(self):
        self.parent = Mock()
        self.resource = Applications("http://api.twilio.com", ("user", "pass"))

    def test_create_application_sms_url_method(self):
        self.resource.create_instance = Mock()
        self.resource.create(sms_method="hey")
        self.resource.create_instance.assert_called_with({"sms_method": "hey"})

    def test_create_application_sms_url(self):
        self.resource.create_instance = Mock()
        self.resource.create(sms_url="hey")
        self.resource.create_instance.assert_called_with({"sms_url": "hey"})

    def test_update_application_sms_url_method(self):
        self.resource.update_instance = Mock()
        self.resource.update("123", sms_method="hey")
        self.resource.update_instance.assert_called_with(
            "123", {"sms_method": "hey"})

    def test_update_application_sms_url(self):
        self.resource.update_instance = Mock()
        self.resource.update("123", sms_url="hey")
        self.resource.update_instance.assert_called_with(
            "123", {"sms_url": "hey"})

    def test_update(self):
        request = Mock()
        request.return_value = (Mock(), {"sid": "123"})
        self.resource.request = request

        self.resource.update("123", voice_url="hey")

        uri = "http://api.twilio.com/Applications/123"
        request.assert_called_with("POST", uri, data={"VoiceUrl": "hey"})

    def test_create(self):
        request = Mock()
        request.return_value = (request, {"sid": "123"})
        request.status_code = 201

        self.resource.request = request

        self.resource.create(friendly_name="hey")

        uri = "http://api.twilio.com/Applications"
        request.assert_called_with("POST", uri, data={"FriendlyName": "hey"})

    @patch("twilio.rest.resources.base.Resource.request")
    def test_delete(self, req):
        """ Deleting an application should work """
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        req.return_value = resp, {}

        app = Application(self.resource, "AP123")
        app.delete()
        uri = "http://api.twilio.com/Applications/AP123"
        req.assert_called_with("DELETE", uri)



# -*- coding: utf-8 -*-
import sys

from six import u

# Backwards compatibility.
from .version import __version__, __version_info__

from .exceptions import TwilioException, TwimlException

from .rest.exceptions import TwilioRestException



from datetime import datetime
from datetime import date

from nose.tools import assert_equal

from twilio.rest.resources import parse_date
from twilio.rest.resources import transform_params
from twilio.rest.resources import convert_keys
from twilio.rest.resources import convert_case
from twilio.rest.resources import convert_boolean
from twilio.rest.resources import normalize_dates


def test_date():
    d = date(2009, 10, 10)
    assert_equal(parse_date(d), "2009-10-10")


def test_datetime():
    d = datetime(2009, 10, 10)
    assert_equal(parse_date(d), "2009-10-10")


def test_string_date():
    d = "2009-10-10"
    assert_equal(parse_date(d), "2009-10-10")


def test_string_date_none():
    d = None
    assert_equal(parse_date(d), None)


def test_string_date_false():
    d = False
    assert_equal(parse_date(d), None)


def test_fparam():
    d = {"HEY": None, "YOU": 3}
    ed = {"YOU": 3}
    assert_equal(transform_params(d), ed)


def test_multi_param():
    d = {"Normal": 3, "Multiple": ["One", "Two"]}
    ed = {"Normal": 3, "Multiple": ["One", "Two"]}
    assert_equal(transform_params(d), ed)


def test_fparam_booleans():
    d = {"HEY": None, "YOU": 3, "Activated": False}
    ed = {"YOU": 3, "Activated": "false"}
    assert_equal(transform_params(d), ed)


def test_normalize_dates():

    @normalize_dates
    def foo(on=None, before=None, after=None):
        return {
            "on": on,
            "before": before,
            "after": after,
        }

    d = foo(on="2009-10-10", before=date(2009, 10, 10),
            after=datetime(2009, 10, 10))

    assert_equal(d["on"], "2009-10-10")
    assert_equal(d["after"], "2009-10-10")
    assert_equal(d["before"], "2009-10-10")


def test_convert_case():
    assert_equal(convert_case("from_"), "From")
    assert_equal(convert_case("to"), "To")
    assert_equal(convert_case("friendly_name"), "FriendlyName")


def test_convert_bool():
    assert_equal(convert_boolean(False), "false")
    assert_equal(convert_boolean(True), "true")
    assert_equal(convert_boolean(1), 1)


def test_convert_keys():
    d = {
        "from_": 0,
        "to": 0,
        "friendly_name": 0,
        "ended": 0,
    }

    ed = {
        "From": 0,
        "To": 0,
        "FriendlyName": 0,
        "EndTime": 0,
    }

    assert_equal(ed, convert_keys(d))



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Message(NextGenInstanceResource):

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this message
        """
        return self.delete_instance()


class Messages(NextGenListResource):

    name = "Messages"
    instance = Message

    def list(self, **kwargs):
        """
        Returns a page of :class:`Message` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Message instance resource representation.

        """
        return self.get_instances(kwargs)

    def create(self, body, **kwargs):
        """
        Create a Message.

        :param str body: The body of the message.
        :param str from: The message author's identity.

        :return: A :class:`Message` object
        """
        kwargs["body"] = body
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given Message
        """
        return self.delete_instance(sid)



from .. import NextGenInstanceResource, NextGenListResource


class Task(NextGenInstanceResource):
    """
    A Task resource.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/tasks>_`
    for more information.

    .. attribute:: sid

        The unique ID for this Task.

    .. attribute:: account_sid

        The ID of the account that owns this Task.

    .. attribute:: workspace_sid

        The ID of the :class:`Workspace` that contains this Task.

    .. attribute:: workflow_sid

        The ID of the :class:`Workflow` that is responsible for routing
        this Task.

    .. attribute:: attributes

        The user-defined JSON string describing the custom attributes of
        this work.

    .. attribute:: age

        The number of seconds since this Task was created.

    .. attribute:: priority

        The current priority score of the task, as assigned by the Workflow.
        Tasks with higher values will be assigned before tasks with lower
        values.

    .. attribute:: task_queue_sid

        The ID of the current TaskQueue this task occupies, controlled by the
        Workflow.

    .. attribute:: assignment_status

        A string representing the assignment state of the task. May be
        'pending', 'reserved', 'assigned', or 'canceled'.

    .. attribute:: reason

        The reason the task was canceled (if applicable).

    .. attribute:: date_created

        The date this task was created, as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The date this task was last updated, as UTC in ISO 8601 format.
    """

    def delete(self):
        """
        Delete a task.
        """
        return self.parent.delete_instance(self.name)

    def update(self, **kwargs):
        """
        Update a task.
        """
        return self.parent.update_instance(self.name, kwargs)


class Tasks(NextGenListResource):
    """ A list of Task resources """

    name = "Tasks"
    instance = Task

    def create(self, attributes, workflow_sid, **kwargs):
        """
        Create a Task.

        :param attributes: Url-encoded JSON string describing the attributes of
            this task. This data will be passed back to the Workflow's
            AssignmentCallbackURL when the Task is assigned to a Worker. An
            example task: { 'task_type': 'call', 'twilio_call_sid': '...',
            'customer_ticket_number': '12345' }.
        :param workflow_sid: The workflow_sid for the Workflow that you would
            like to handle routing for this Task.
        :param timeout: If provided, time-to-live for the task in seconds,
            before it is automatically canceled
        """
        kwargs['attributes'] = attributes
        kwargs['workflow_sid'] = workflow_sid
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete the given task
        """
        return self.delete_instance(sid)

    def update(self, sid, **kwargs):
        """
        Update a :class:`Task` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)



import unittest

from mock import patch, Mock, sentinel, ANY
from nose.tools import assert_equal, assert_true

from twilio.rest.resources.imports import json
from twilio.rest import TwilioRestClient, resources, TwilioTaskRouterClient
from tests.tools import create_mock_json

AUTH = ("ACCOUNT_SID", "AUTH_TOKEN")


class RestClientTest(unittest.TestCase):
    def setUp(self):
        self.client = TwilioRestClient("ACCOUNT_SID", "AUTH_TOKEN")
        self.task_router_client = TwilioTaskRouterClient("ACCOUNT_SID",
                                                         "AUTH_TOKEN")

    @patch("twilio.rest.base.make_request")
    def test_request(self, mock):
        self.client.request("2010-04-01", method="GET")
        mock.assert_called_with("GET", "https://api.twilio.com/2010-04-01",
                                headers={"User-Agent": ANY,
                                         'Accept-Charset': 'utf-8'},
                                params={}, auth=AUTH, data=None)
        called_kwargs = mock.mock_calls[0][2]
        self.assertTrue(
            'twilio-python' in called_kwargs['headers']['User-Agent']
        )

    def test_connect_apps(self):
        assert_true(isinstance(self.client.connect_apps,
                               resources.ConnectApps))

    def test_authorized_apps(self):
        assert_true(isinstance(self.client.authorized_connect_apps,
                               resources.AuthorizedConnectApps))

    @patch("twilio.rest.resources.base.make_request")
    def test_conferences(self, mock):
        mock.return_value = Mock()
        mock.return_value.ok = True
        mock.return_value.content = '{"conferences": []}'
        self.client.conferences.list()

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_members(self, mock):
        resp = create_mock_json("tests/resources/members_list.json")
        mock.return_value = resp
        self.client.members("QU123").list()
        uri = "https://api.twilio.com/2010-04-01/Accounts/ACCOUNT_SID" \
              "/Queues/QU123/Members"
        mock.assert_called_with("GET", uri, params={}, auth=AUTH,
                                use_json_extension=True)

    @patch("twilio.rest.resources.base.make_request")
    def test_workflows(self, request):
        resp = create_mock_json(
            "tests/resources/task_router/workflows_list.json"
        )
        request.return_value = resp
        workflows = self.task_router_client.workflows("WS123")
        workflows = workflows.list()
        assert_true(workflows[0].sid is not None)
        uri = "https://taskrouter.twilio.com/v1/Workspaces/WS123/Workflows"
        request.assert_called_with("GET", uri, headers=ANY, params={},
                                   auth=AUTH)


class RestClientTimeoutTest(unittest.TestCase):
    def setUp(self):
        self.client = TwilioRestClient("ACCOUNT_SID", "AUTH_TOKEN",
                                       timeout=sentinel.timeout)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_members(self, mock_request):
        resp = create_mock_json("tests/resources/members_list.json")
        mock_request.return_value = resp
        self.client.members("QU123").list()
        mock_request.assert_called_with("GET", ANY, params=ANY, auth=AUTH,
                                        timeout=sentinel.timeout,
                                        use_json_extension=True)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_arbitrary_member(self, mock_request):
        mock_response = Mock()
        mock_response.ok = True
        mock_response.content = json.dumps({"short_codes": []})
        mock_request.return_value = mock_response
        assert_equal([], self.client.sms.short_codes.list())
        mock_request.assert_called_once_with("GET", ANY, params=ANY, auth=AUTH,
                                             timeout=sentinel.timeout,
                                             use_json_extension=True)



import unittest

from mock import Mock
from nose.tools import assert_equal, assert_true

from twilio.rest.exceptions import TwilioException
from twilio.rest.resources import AvailablePhoneNumber
from twilio.rest.resources import AvailablePhoneNumbers
from twilio.rest.resources import PhoneNumbers
from twilio.rest.resources import UNSET_TIMEOUT


class AvailablePhoneNumberTest(unittest.TestCase):

    def setUp(self):
        self.parent = Mock()
        self.instance = AvailablePhoneNumber(self.parent)

    def test_init(self):
        assert_equal(self.instance.name, "")

    def test_purchase(self):
        self.instance.phone_number = "+123"
        self.instance.purchase(voice_url="http://www.google.com")

        self.parent.purchase.assert_called_with(
            voice_url="http://www.google.com",
            phone_number="+123")


class AvailablePhoneNumbersTest(unittest.TestCase):

    def setUp(self):
        self.resource = AvailablePhoneNumbers("http://api.twilio.com",
                                              ("user", "pass"), UNSET_TIMEOUT, Mock())

    def test_get(self):
        self.assertRaises(TwilioException, self.resource.get, "PN123")

    def test_list(self):
        request = Mock()
        request.return_value = (Mock(), {"available_phone_numbers": []})
        self.resource.request = request

        self.resource.list()

        uri = "http://api.twilio.com/AvailablePhoneNumbers/US/Local"
        request.assert_called_with("GET", uri, params={})

    def test_load_instance(self):
        instance = self.resource.load_instance({"hey": "you"})
        assert_true(isinstance(instance.parent, Mock))
        assert_equal(instance.hey, "you")

    def test_purchase_status_callback(self):
        request = Mock()
        request.return_value = (Mock(), {"available_phone_numbers": []})
        self.resource.request = request

        self.resource.list()

        uri = "http://api.twilio.com/AvailablePhoneNumbers/US/Local"
        request.assert_called_with("GET", uri, params={})

    def test_mobile(self):
        request = Mock()
        request.return_value = (Mock(), {"available_phone_numbers": []})
        self.resource.request = request

        self.resource.list(type='mobile', country='GB')

        uri = "http://api.twilio.com/AvailablePhoneNumbers/GB/Mobile"
        request.assert_called_with("GET", uri, params={})


class PhoneNumbersTest(unittest.TestCase):

    def setUp(self):
        self.resource = PhoneNumbers("http://api.twilio.com",
                                     ("user", "pass"))

    def test_reference(self):
        assert_equal(self.resource.available_phone_numbers.phone_numbers,
                     self.resource)

    def test_purchase_status_callback(self):
        request = Mock()
        response = Mock()
        response.status_code = 201
        request.return_value = (response, {"sid": ""})
        self.resource.request = request

        self.resource.purchase(area_code="530", status_callback_url="http://",
                               status_callback_method="POST")

        uri = "http://api.twilio.com/IncomingPhoneNumbers"

        data = {
            "AreaCode": "530",
            "StatusCallback": "http://",
            "StatusCallbackMethod": "POST",
            }

        request.assert_called_with("POST", uri, data=data)



class WorkflowRuleTarget:
    """
    Workflow Rule target which is encompassed
    inside targets

    ..attribute::queue

      The queue which will handle the task matching this filter target

    ..attribute::expression

       The dynamic expression if any for this matching

    ..attribute::priority

      The priority for the target

    ..attribute::timeout

      The timeout before the reservation expires.
    """
    def __init__(self, queue, expression, priority, timeout):

        self.queue = queue
        self.expression = expression
        self.priority = priority
        self.timeout = timeout



import unittest

from mock import Mock, patch

from twilio.rest.resources import ConnectApps


class ConnectAppTest(unittest.TestCase):

    def setUp(self):
        self.parent = Mock()
        self.uri = "/base"
        self.auth = ("AC123", "token")
        self.resource = ConnectApps(self.uri, self.auth)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_get(self, mock):
        mock.return_value = Mock()
        mock.return_value.content = '{"sid": "SID"}'

        self.resource.get("SID")
        mock.assert_called_with("GET", "/base/ConnectApps/SID",
            auth=self.auth, use_json_extension=True)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_list_with_paging(self, mock):
        mock.return_value = Mock()
        mock.return_value.content = '{"connect_apps": []}'

        self.resource.list(page=1, page_size=50)
        mock.assert_called_with("GET", "/base/ConnectApps",
                                params={"Page": 1, "PageSize": 50},
                                auth=self.auth,
                                use_json_extension=True)

    @patch("twilio.rest.resources.base.make_twilio_request")
    def test_list(self, mock):
        mock.return_value = Mock()
        mock.return_value.content = '{"connect_apps": []}'

        self.resource.list()
        mock.assert_called_with("GET", "/base/ConnectApps",
                                params={}, auth=self.auth,
                                use_json_extension=True)

    def test_create(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'create')

    def test_delete(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'delete')

    def test_update(self):
        self.assertRaises(AttributeError, getattr, self.resource, 'update')



from datetime import date
import unittest

from mock import Mock
from twilio.rest.resources import Conferences

DEFAULT = {
    'DateUpdated<': None,
    'DateUpdated>': None,
    'DateUpdated': None,
    'DateCreated<': None,
    'DateCreated>': None,
    'DateCreated': None,
}


class ConferenceTest(unittest.TestCase):

    def setUp(self):
        self.resource = Conferences("foo", ("sid", "token"))
        self.params = DEFAULT.copy()

    def test_list(self):
        self.resource.get_instances = Mock()
        self.resource.list()
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_after(self):
        self.resource.get_instances = Mock()
        self.resource.list(created_after=date(2011, 1, 1))
        self.params["DateCreated>"] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_on(self):
        self.resource.get_instances = Mock()
        self.resource.list(created=date(2011, 1, 1))
        self.params["DateCreated"] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)

    def test_list_before(self):
        self.resource.get_instances = Mock()
        self.resource.list(created_before=date(2011, 1, 1))
        self.params["DateCreated<"] = "2011-01-01"
        self.resource.get_instances.assert_called_with(self.params)



# parse_qs
try:
    from urlparse import parse_qs
except ImportError:
    from cgi import parse_qs

# json
try:
    import json
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        from django.utils import simplejson as json

# httplib2
import httplib2

# socks
try:
    from httplib2 import socks
    from httplib2.socks import (
        PROXY_TYPE_HTTP,
        PROXY_TYPE_SOCKS4,
        PROXY_TYPE_SOCKS5
    )
except ImportError:
    import socks
    from socks import (
        PROXY_TYPE_HTTP,
        PROXY_TYPE_SOCKS4,
        PROXY_TYPE_SOCKS5
    )



import unittest

from mock import Mock
from nose.tools import assert_equal, raises

from twilio.exceptions import TwilioException
from twilio.rest.resources import Addresses, DependentPhoneNumbers


class AddressesTest(unittest.TestCase):
    def setUp(self):
        self.parent = Mock()
        self.resource = Addresses("http://api.twilio.com", ("user", "pass"))

    def test_update(self):
        request = Mock()
        request.return_value = (Mock(), {"sid": "123"})
        self.resource.request = request

        self.resource.update("123", friendly_name="hi")

        uri = "http://api.twilio.com/Addresses/123"
        request.assert_called_with("POST", uri, data={"FriendlyName": "hi"})

    @raises(TwilioException)
    def test_update_rejects_iso_country(self):
        self.resource.update("123", iso_country="CA")

    def test_dependent_phone_numbers(self):
        pn_list = DependentPhoneNumbers(
            'http://api.twilio.com/mock',
            ('user', 'pass'),
        )
        request = Mock()
        request.return_value = (
            Mock(),
            {
                "dependent_phone_numbers": [{"sid": "PN123"}],
                "total": 1,
                "page": 0,
                "page_size": 50,
            },
        )
        pn_list.request = request

        result = pn_list.list()
        request.assert_called_with(
            "GET",
            "http://api.twilio.com/mock/DependentPhoneNumbers",
            params={},
        )
        assert_equal(result[0].sid, 'PN123')



from .. import NextGenInstanceResource, NextGenListResource


class PhoneNumbers(object):
    """Holds references to the Number pricing resources."""

    name = "Number"
    key = "Number"

    def __init__(self, base_uri, auth, timeout):
        self.uri = "%s/PhoneNumbers" % base_uri
        self.countries = PhoneNumberCountries(self.uri, auth, timeout)


class PhoneNumberCountry(NextGenInstanceResource):
    """Pricing information for Twilio Phone Numbers in a specific country.

    Twilio numbers are billed monthly, and the prices returned reflect
    current Twilio list pricing before and after any discounts available for
    the requesting account are applied.

    .. attribute:: country

        The full name of the country.

    .. attribute:: iso_country

        The country's 2-character ISO 3166-1 code.

    .. attribute:: price_unit

        The currency in which prices are measured, in ISO 4127 format
        (e.g. 'usd', 'eur', 'jpy').

    .. attribute:: phone_number_prices

        A list of dicts containing pricing information as follows:
            - type: "local", "mobile", "national", or "toll_free"
            - base_price: the base price per month for this Twilio number type
            - current_price: the current price per month (including discounts)
            for this Twilio number type
    """

    id_key = "iso_country"


class PhoneNumberCountries(NextGenListResource):
    """A list of countries where Twilio Phone Numbers are available.

    The returned list of PhoneNumberCountry objects will not have pricing
    information populated. To get pricing information for a specific country,
    retrieve it with the :meth:`get` method.
    """

    instance = PhoneNumberCountry
    key = "countries"
    name = "Countries"

    def get(self, iso_country):
        """Retrieve pricing information for Twilio Numbers in the specified
        country.

        :param iso_country: The two-letter ISO code for the country
        """
        return self.get_instance(iso_country)

    def list(self):
        """Retrieve the list of countries in which Twilio Numbers are
        available."""

        resp, page = self.request("GET", self.uri)

        return [self.load_instance(i) for i in page[self.key]]



from . import (
    ListResource,
    InstanceResource,
    transform_params,
)


class CallFeedback(InstanceResource):

    def __init__(self, parent):
        self.parent = parent
        super(CallFeedback, self).__init__(
            parent,
            None,
        )


class CallFeedbackFactory(ListResource):

    """
    CallFeedback is a unique endpoint in the API in that it only
    has an instance representation, and that instance representation
    lives at the same URL you POST to to create it. Here, our
    ListResource class acts as a Factory resource.
    """

    name = "Feedback"
    instance = CallFeedback

    def create(self, **kwargs):
        """
        Create a :class:`CallFeedback` object for the parent call.

        :param int quality: The score quality. Must be an
            int between 1 and 5.
        :param list issue: A list of issues. The issue types are
            found at the CallFeedback rest docs.
        """
        return self.create_instance(kwargs)

    def get(self, **kwargs):
        """ Get the feedback for this call

        Usage:

        .. code-block:: python

            feedback = client.calls.get("CA123").feedback
            print feedback.issues

        :rtype: :class:`~twilio.rest.resources.InstanceResource`
        :raises: a :exc:`~twilio.TwilioRestException` if the request fails
        """
        params = transform_params(kwargs)
        _, data = self.request("GET", self.uri, params=params)
        return self.load_instance(data)

    def load_instance(self, data):
        # Overridden because CallFeedback instances
        # don't contain sids
        instance = self.instance(self)
        instance.load(data)
        return instance


class CallFeedbackSummaryInstance(InstanceResource):

    def __init__(self, parent):
        self.parent = parent
        super(CallFeedbackSummaryInstance, self).__init__(
            parent,
            None,
        )


class CallFeedbackSummary(ListResource):

    name = "Summary"
    key = "Feedback"
    instance = CallFeedbackSummaryInstance

    def __init__(self, parent, *args, **kwargs):
        super(CallFeedbackSummary, self).__init__(*args, **kwargs)
        self.base_uri = parent.uri

    def get(self, **kwargs):
        """ Get the feedback summary for calls on this account

        Usage:

        .. code-block:: python

            summary = client.calls.summary.get()
            print summary.quality_score_average

        :rtype: :class:`~twilio.rest.resources.InstanceResource`
        :raises: a :exc:`~twilio.TwilioRestException` if the request fails
        """
        params = transform_params(kwargs)
        _, data = self.request('GET', self.uri, params=params)
        return self.load_instance(data)

    def load_instance(self, data):
        # Overridden because CallFeedback summaries
        # do not contain sids
        instance = self.instance(self)
        instance.load(data)
        return instance



from .util import (
    transform_params, format_name, parse_date, convert_boolean, convert_case,
    convert_keys, normalize_dates, UNSET_TIMEOUT
)
from .base import (
    Response, Resource, InstanceResource, ListResource,
    NextGenInstanceResource, NextGenListResource,
    make_request, make_twilio_request
)
from .phone_numbers import (
    AvailablePhoneNumber, AvailablePhoneNumbers, PhoneNumber, PhoneNumbers
)
from .recordings import Recording, Recordings
from .transcriptions import Transcription, Transcriptions

from .notifications import Notification, Notifications
from .connect_apps import (
    ConnectApp, ConnectApps, AuthorizedConnectApp, AuthorizedConnectApps
)
from .calls import Call, Calls
from .caller_ids import CallerIds, CallerId
from .call_feedback import (
    CallFeedbackFactory, CallFeedback, CallFeedbackSummary,
    CallFeedbackSummaryInstance
)
from .connection import Connection
from .sandboxes import Sandbox, Sandboxes
from .sms_messages import (
    Sms, SmsMessage, SmsMessages, ShortCode, ShortCodes)
from .conferences import (
    Participant, Participants, Conference, Conferences
)
from .queues import (
    Member, Members, Queue, Queues,
)
from .applications import (
    Application, Applications
)
from .accounts import Account, Accounts

from .usage import Usage

from .messages import Message, Messages

from .media import Media, MediaList

from .signing_keys import SigningKey, SigningKeys

from .sip import Sip

from .task_router import (
    Activity,
    Activities,
    Event,
    Events,
    Reservation,
    Reservations,
    Task,
    Tasks,
    TaskQueue,
    TaskQueues,
    Worker,
    Workers,
    Workflow,
    Workflows,
    Workspace,
    Workspaces,
)

from .tokens import Token, Tokens

from .addresses import (
    Address,
    Addresses,
    DependentPhoneNumber,
    DependentPhoneNumbers,
)

from .keys import Key, Keys



from mock import patch, Mock
from twilio.rest.resources.ip_messaging import Users, User
from tests.tools import create_mock_json

BASE_URI = "https://ip-messaging.twilio.com/v1/Services/ISxxx"
ACCOUNT_SID = "ACaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
AUTH = (ACCOUNT_SID, "token")
USER_SID = "USaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"

list_resource = Users(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_user(mock):
    resp = create_mock_json("tests/resources/ip_messaging/user_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Users" % (BASE_URI)
    list_resource.create('test_id')
    exp_params = {
        'Id': "test_id"
    }

    mock.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                            use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/ip_messaging/user_instance.json")
    mock.return_value = resp

    uri = "%s/Users/%s" % (BASE_URI, USER_SID)
    list_resource.get(USER_SID)

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=False)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete(req):
    """ Deleting a call should work """
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    req.return_value = resp, {}

    app = User(list_resource, "US123")
    app.delete()
    uri = "%s/Users/US123" % (BASE_URI)
    req.assert_called_with("DELETE", uri)



from mock import patch, Mock
from nose.tools import raises

from tests.tools import create_mock_json
from twilio.rest.resources import Usage
from twilio.rest.resources.usage import UsageTriggers, UsageTrigger

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")

usage = Usage(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_triggers_create(request):
    resp = create_mock_json("tests/resources/usage_triggers_instance.json")
    resp.status_code = 201
    request.return_value = resp

    usage.triggers.create(
        friendly_name="foo",
        usage_category="sms",
        trigger_by="count",
        recurring="price",
        trigger_value="10.00",
        callback_url="http://www.example.com",
        callback_method="POST"
    )

    uri = "%s/Usage/Triggers" % BASE_URI
    request.assert_called_with("POST", uri, data={
        "FriendlyName": "foo",
        "UsageCategory": "sms",
        "TriggerBy": "count",
        "Recurring": "price",
        "TriggerValue": "10.00",
        "CallbackUrl": "http://www.example.com",
        "CallbackMethod": "POST"
    }, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_triggers_paging(request):
    resp = create_mock_json("tests/resources/usage_triggers_list.json")
    request.return_value = resp

    uri = "%s/Usage/Triggers" % BASE_URI
    usage.triggers.list(
        recurring="daily",
        usage_category="sms",
        trigger_by="count")

    request.assert_called_with("GET", uri, params={
        "Recurring": "daily",
        "UsageCategory": "sms",
        "TriggerBy": "count"
    }, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_records_paging(request):
    resp = create_mock_json("tests/resources/usage_records_list.json")
    request.return_value = resp

    uri = "%s/Usage/Records" % BASE_URI
    usage.records.list(
        start_date="2012-10-12",
        end_date="2012-10-13",
        category="sms")

    request.assert_called_with("GET", uri, params={
        "StartDate": "2012-10-12",
        "EndDate": "2012-10-13",
        "Category": "sms"
    }, auth=AUTH, use_json_extension=True)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete_trigger(req):
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    req.return_value = resp, {}

    triggers = UsageTriggers("https://api.twilio.com", None)
    trigger = UsageTrigger(triggers, "UT123")
    trigger.delete()
    uri = "https://api.twilio.com/Usage/Triggers/UT123"
    req.assert_called_with("DELETE", uri)


@raises(AttributeError)
def test_records_create():
    usage.records.all.create


@raises(AttributeError)
def test_records_delete():
    usage.records.all.delete


@raises(AttributeError)
def test_records_get():
    usage.records.all.get('abc')



import unittest

from mock import patch, Mock

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.workflows import Workflows, Workflow


AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
WORKFLOW_SID = "WFaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class WorkflowTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/workflows_instance.json')
        resp.status_code = 201
        request.return_value = resp

        workflows = Workflows(BASE_URI, AUTH)
        workflows.create("Test Workflow", "configuration", "http://www.example.com")
        exp_params = {
            'FriendlyName': "Test Workflow",
            'Configuration': "configuration",
            'AssignmentCallbackUrl': 'http://www.example.com'
        }

        request.assert_called_with("POST", "{0}/Workflows".format(BASE_URI), data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workflows/{1}".format(BASE_URI, WORKFLOW_SID)
        list_resource = Workflows(BASE_URI, AUTH)
        workflow = Workflow(list_resource, WORKFLOW_SID)
        workflow.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/Workflows/{1}".format(BASE_URI, WORKFLOW_SID)
        list_resource = Workflows(BASE_URI, AUTH)
        list_resource.delete(WORKFLOW_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/workflows_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workflows/{1}".format(BASE_URI, WORKFLOW_SID)
        list_resource = Workflows(BASE_URI, AUTH)
        list_resource.get(WORKFLOW_SID)
        request.assert_called_with("GET", uri, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workflows_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Workflows".format(BASE_URI)
        list_resource = Workflows(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/workflows_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workflows/{1}".format(BASE_URI, WORKFLOW_SID)
        list_resource = Workflows(BASE_URI, AUTH)
        workflow = Workflow(list_resource, WORKFLOW_SID)
        workflow.update(friendly_name='Test Workflow', configuration='configuration',
                        assignment_callback_url='http://www.example.com')
        exp_params = {
            'FriendlyName': "Test Workflow",
            'Configuration': "configuration",
            'AssignmentCallbackUrl': 'http://www.example.com'
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/workflows_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/Workflows/{1}".format(BASE_URI, WORKFLOW_SID)
        list_resource = Workflows(BASE_URI, AUTH)
        list_resource.update(WORKFLOW_SID, friendly_name='Test Workflow', configuration='configuration',
                             assignment_callback_url='http://www.example.com')
        exp_params = {
            'FriendlyName': "Test Workflow",
            'Configuration': "configuration",
            'AssignmentCallbackUrl': 'http://www.example.com'
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   use_json_extension=False)



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Credential(NextGenInstanceResource):

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this credential
        """
        return self.delete_instance()


class Credentials(NextGenListResource):

    name = "Credentials"
    instance = Credential

    def list(self, **kwargs):
        """
        Returns a page of :class:`Credential` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Credential instance resource representation.

        :param date after: Only list alerts logged after this datetime
        :param date before: Only list alerts logger before this datetime
        :param log_level: If 'error', only shows errors. If 'warning',
         only show warnings
        """
        return self.get_instances(kwargs)

    def create(self, type, **kwargs):
        """
        Make a phone call to a number.

        :param str type: The type of credential
        :param str friendly_name: The friendly name of the credential.

        :return: A :class:`Credential` object
        """
        kwargs["type"] = type
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given Credential
        """
        return self.delete_instance(sid)



from mock import patch

from nose.tools import assert_equal

from tests.tools import create_mock_json
from twilio.rest.resources.lookups.phone_numbers import PhoneNumbers


AUTH = ('AC123', 'foobar')
TIMEOUT = 30


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get_phone_number(request):
    resp = create_mock_json(
        "tests/resources/lookups/phone_number_instance.json",
    )
    request.return_value = resp

    phone_numbers = PhoneNumbers('/v1', AUTH, TIMEOUT)
    pn = phone_numbers.get('+15108675309')
    assert_equal(pn.phone_number, '+15108675309')
    request.assert_called_with('GET', '/v1/PhoneNumbers/+15108675309',
                               auth=AUTH, timeout=TIMEOUT, params={},
                               use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get_carrier_info(request):
    resp = create_mock_json(
        "tests/resources/lookups/phone_number_instance.json",
    )
    request.return_value = resp

    phone_numbers = PhoneNumbers('/v1', AUTH, TIMEOUT)
    pn = phone_numbers.get('+15108675309', include_carrier_info=True)
    assert_equal(pn.phone_number, '+15108675309')
    request.assert_called_with('GET', '/v1/PhoneNumbers/+15108675309',
                               auth=AUTH, timeout=TIMEOUT,
                               params={'Type': 'carrier'},
                               use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get_with_country_code(request):
    resp = create_mock_json(
        "tests/resources/lookups/phone_number_instance.json",
    )
    request.return_value = resp

    phone_numbers = PhoneNumbers('/v1', AUTH, TIMEOUT)
    pn = phone_numbers.get('510-867-5309', country_code='US')
    assert_equal(pn.phone_number, '+15108675309')
    request.assert_called_with('GET', '/v1/PhoneNumbers/510-867-5309',
                               auth=AUTH, timeout=TIMEOUT,
                               params={'CountryCode': 'US'},
                               use_json_extension=False)



from .. import NextGenInstanceResource, NextGenListResource


class Voice(object):
    """Holds references to the Voice pricing resources."""

    name = "Voice"
    key = "voice"

    def __init__(self, base_uri, auth, timeout):
        self.uri = "%s/Voice" % base_uri
        self.countries = VoiceCountries(self.uri, auth, timeout)
        self.numbers = VoiceNumbers(self.uri, auth, timeout)


class VoiceCountry(NextGenInstanceResource):
    """Pricing information for Twilio Voice services in a specific country.

    .. attribute:: country

        The full name of the country.

    .. attribute:: iso_country

        The country's 2-character ISO code.

    .. attribute:: price_unit

        The currency in which prices are measured, in ISO 4127 format
        (e.g. 'usd', 'eur', 'jpy').

    .. attribute:: outbound_prefix_prices

        A list of dicts containing pricing information as follows:
            - prefix_list: a list of number prefixes in the requested country
              that have the same pricing
            - friendly_name: a descriptive name for this prefix set
            - call_base_price: the base price per minute for calls to numbers
              matching any of these prefixes
            - call_current_price: the current price per minute (including
              volume discounts, etc.) for your account to make calls to
              numbers matching these prefixes

    .. attribute:: inbound_call_prices

        A list of dicts containing pricing information for inbound calls:
            - number_type: 'local', 'mobile', 'national', or 'toll_free'
            - call_base_price: the base price per minute to receive a call
              to this number type
            - call_current_price: the current price per minute (including
              volume discounts, etc.) for your account to receive a call
              to this number type
    """

    id_key = "iso_country"


class VoiceCountries(NextGenListResource):

    instance = VoiceCountry
    key = "countries"
    name = "Countries"

    def get(self, iso_country):
        """Retrieve pricing information for Twilio Voice in the specified
        country.

        :param iso_country: The two-letter ISO code for the country
        """
        return self.get_instance(iso_country)

    def list(self):
        """Retrieve the list of countries in which Twilio Voice is
        available."""
        resp, page = self.request("GET", self.uri)

        return [self.load_instance(i) for i in page[self.key]]


class VoiceNumber(NextGenInstanceResource):
    """Pricing information for Twilio Voice services to and from a given
    phone number.

    .. attribute:: phone_number

        The E.164-formatted phone number this pricing information applies to

    .. attribute:: country

        The name of the country this phone number belongs to

    .. attribute:: iso_country

        The two-character ISO code for the country

    .. attribute:: outbound_call_price

        A dict containing pricing information for outbound calls to this
        number:
            - base_price: the base price per minute for a call to this number
            - current_price: the current price per minute (including discounts,
            etc.) for a call to this number

    .. attribute:: inbound_call_price

        A dict containing pricing information for inbound call to this number,
        or null if this number is not Twilio-hosted.

            - number_type: "local", "mobile", "national", or "toll_free"
            - call_base_price: the base price per minute to receive a call to
            this number
            - call_current_price: the current price per minute (including
            discounts, etc.) to receive a call to this number
    """

    id_key = "number"


class VoiceNumbers(NextGenListResource):

    instance = VoiceNumber
    key = "numbers"
    name = "Numbers"

    def get(self, phone_number):
        """ Retrieve pricing information for a specific phone number.
        :param phone_number: the E.164-formatted number to retrieve info for
        :return: a :class:`VoiceNumber` instance
        """

        return self.get_instance(phone_number)



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class ConversationsRoot(object):

    def __init__(self, base_uri, *args, **kwargs):
        self.uri = "%s/Conversations" % base_uri
        self._instance_base = Conversations(self.uri, *args, **kwargs)
        self.in_progress = Conversations("%s/InProgress" % self.uri, *args,
                                         **kwargs)
        self.completed = Conversations("%s/Completed" % self.uri, *args,
                                       **kwargs)

    def get(self, sid):
        return self._instance_base.get(sid)

    def delete_instance(self, sid):
        return self._instance_base.delete_instance(sid)


class Conversation(NextGenInstanceResource):
    """A Conversation instance representing a call
       between two or more participants.

    .. attribute:: sid

    .. attribute:: account_sid

    .. attribute:: status

    .. attribute:: date_created

    .. attribute:: start_time

    .. attribute:: end_time

    .. attribute:: duration

    .. attribute:: url
    """
    pass


class Conversations(NextGenListResource):

    name = "Conversations"
    instance = Conversation

    def __init__(self, uri, *args, **kwargs):
        super(Conversations, self).__init__(uri, *args, **kwargs)
        # This list is exposed at two different locations: /InProgress
        # and /Completed. The parent Root object will hand us the full URL
        # to set up at.
        self._uri = uri

    @property
    def uri(self):
        return self._uri



from .util import normalize_dates, parse_date
from . import InstanceResource, ListResource


class ShortCode(InstanceResource):

    def update(self, **kwargs):
        return self.parent.update(self.name, **kwargs)


class ShortCodes(ListResource):

    name = "ShortCodes"
    key = "short_codes"
    instance = ShortCode

    def list(self, **kwargs):
        """
        Returns a page of :class:`ShortCode` resources as a list. For
        paging information see :class:`ListResource`.

        :param short_code: Only show the ShortCode resources that match this
                           pattern. You can specify partial numbers and use '*'
                           as a wildcard for any digit.
        :param friendly_name: Only show the ShortCode resources with friendly
                              names that exactly match this name.
        """
        return self.get_instances(kwargs)

    def update(self, sid, url=None, method=None, fallback_url=None,
               fallback_method=None, **kwargs):
        """
        Update a specific :class:`ShortCode`, by specifying the sid.

        :param friendly_name: Description of the short code, with maximum
                              length 64 characters.
        :param api_version: SMSs to this short code will start a new TwiML
                            session with this API version.
        :param url: The URL that Twilio should request when somebody sends an
                    SMS to the short code.
        :param method: The HTTP method that should be used to request the url.
        :param fallback_url: A URL that Twilio will request if an error occurs
                             requesting or executing the TwiML at the url.
        :param fallback_method: The HTTP method that should be used to request
                                the fallback_url.
        """
        kwargs["sms_url"] = kwargs.get("sms_url", url)
        kwargs["sms_method"] = kwargs.get("sms_method", method)
        kwargs["sms_fallback_url"] = \
            kwargs.get("sms_fallback_url", fallback_url)
        kwargs["sms_fallback_method"] = \
            kwargs.get("sms_fallback_method", fallback_method)
        return self.update_instance(sid, kwargs)


class Sms(object):
    """
    Holds all the specific SMS list resources
    """

    name = "SMS"
    key = "sms"

    def __init__(self, base_uri, auth, timeout):
        self.uri = "%s/SMS" % base_uri
        self.messages = SmsMessages(self.uri, auth, timeout)
        self.short_codes = ShortCodes(self.uri, auth, timeout)


class SmsMessage(InstanceResource):
    """ An instance of an SMS Message

   .. attribute:: sid

      A 34 character string that uniquely identifies this resource.

   .. attribute:: date_created

      The date that this resource was created, given in RFC 2822 format.

   .. attribute:: date_updated

      The date that this resource was last updated, given in RFC 2822 format.

   .. attribute:: date_sent

      The date that the SMS was sent, given in RFC 2822 format.

   .. attribute:: account_sid

      The unique id of the Account that sent this SMS message.

   .. attribute:: from

      The phone number that initiated the message in E.164 format.
      For incoming messages, this will be the remote phone.
      For outgoing messages, this will be one of your Twilio phone numbers.

   .. attribute:: to

      The phone number that received the message in E.164 format.
      For incoming messages, this will be one of your Twilio phone numbers.
      For outgoing messages, this will be the remote phone.

   .. attribute:: body

      The text body of the SMS message.

   .. attribute:: status

      The status of this SMS message. Either queued, sending, sent, or failed.

   .. attribute:: direction

        The direction of this SMS message. ``incoming`` for incoming
        messages, ``outbound-api`` for messages initiated via the REST
        API, ``outbound-call`` for messages initiated during a call or
        ``outbound-reply`` for messages initiated in response to an incoming
        SMS.

   .. attribute:: price

      The amount billed for the message.

   .. attribute:: api_version

      The version of the Twilio API used to process the SMS message.

   .. attribute:: uri

      The URI for this resource, relative to https://api.twilio.com
    """

    pass


class SmsMessages(ListResource):

    name = "Messages"
    key = "sms_messages"
    instance = SmsMessage

    def create(self, from_=None, **kwargs):
        """
        Create and send a SMS Message.

        :param str to: The destination phone number.
        :param str `from_`: The phone number sending this message
            (must be a verified Twilio number)
        :param str body: The message you want to send,
            limited to 160 characters.
        :param status_callback: A URL that Twilio will POST to when
            your message is processed.
        :param str application_sid: The 34 character sid of the application
            Twilio should use to handle this phone call.

        Usage:

        .. code-block::python

            message = client.sms.messages.create(to="+12316851234",
                from_="+15555555555",
                body="Hello there!")

        """
        kwargs["from"] = from_
        return self.create_instance(kwargs)

    @normalize_dates
    def list(self, from_=None, before=None, after=None, date_sent=None, **kw):
        """
        Returns a page of :class:`~twilio.rest.resources.SmsMessage` resources
        as a list. For paging information see :class:`ListResource`.

        :param to: Only show SMS messages to this phone number.
        :param from_: Only show SMS messages from this phone number.
        :param date after: Only list SMS messages sent after this date.
        :param date before: Only list SMS message sent before this date.
        :param date date_sent: Only list SMS message sent on this date.
        :param `from_`: Only show SMS messages from this phone number.
        :param date after: Only list SMS messages logged after this datetime
        :param date before: Only list SMS messages logged before this datetime
        """
        kw["From"] = from_
        kw["DateSent<"] = before
        kw["DateSent>"] = after
        kw["DateSent"] = parse_date(date_sent)
        return self.get_instances(kw)



from . import InstanceResource, ListResource


class Token(InstanceResource):
    """ A Token instance.

    .. attribute:: username

       The temporary username that uniquely identifies a Token.

    .. attribute:: password

       The temporary password that the username will use when
       authenticating with Twilio.

    .. attribute:: ttl

       The duration in seconds for which the username and password
       are valid, the default value is 86,400 (24 hours).

    .. attribute:: account_sid

       The unique id of the Account that created this Token.

    .. attribute:: ice_servers

       An array representing the ephemeral credentials and the
       STUN and TURN server URIs.

    .. attribute:: date_created

       The date that this resource was created, given in RFC 2822 format.

    .. attribute:: date_updated

       The date that this resource was last updated, given in RFC 2822 format.
    """
    id_key = "username"


class Tokens(ListResource):
    name = "Tokens"
    key = "tokens"
    instance = Token

    def create(self, ttl=None, **kwargs):
        """
        Create a new Token.

        :param int ttl: The duration in seconds for which the token
            is valid, the default value is 86,400 (24 hours).
        """
        kwargs['ttl'] = ttl
        return self.create_instance(kwargs)



from .. import NextGenInstanceResource, NextGenListResource, transform_params


class StatisticsInstance(NextGenInstanceResource):
    """
    A resource representing collected statistics about TaskRouter
    resources.

    See the `TaskRouter API reference
    <https://www.twilio.com/docs/taskrouter/statistics>_`
    for more information.
    """
    def __init__(self, parent):
        self.parent = parent
        super(StatisticsInstance, self).__init__(
            parent,
            None,
        )


class Statistics(NextGenListResource):
    name = 'Statistics'
    key = 'statistics'
    instance = StatisticsInstance

    def get(self, **kwargs):
        params = transform_params(kwargs)
        _, data = self.request('GET', self.uri, params=params)
        return self.load_instance(data)

    def load_instance(self, data):
        # Overridden because Statistics instances
        # don't contain sids
        instance = self.instance(self)
        instance.load(data)
        return instance



from twilio.rest.base import TwilioClient
from twilio.rest.resources import UNSET_TIMEOUT
from twilio.rest.resources.pricing import (
    PhoneNumbers,
    Voice,
)


class TwilioPricingClient(TwilioClient):
    """
    A client for accessing the Twilio Pricing API.

    :param str account: Your Account SID from `your dashboard
        <https://twilio.com/user/account>`_
    :param str token: Your Auth Token from `your dashboard
        <https://twilio.com/user_account>`_
    :param float timeout: The socket connect and read timeout for requests
    to Twilio
    """

    def __init__(self, account=None, token=None,
                 base="https://pricing.twilio.com", version="v1",
                 timeout=UNSET_TIMEOUT):
        super(TwilioPricingClient, self).__init__(account, token, base,
                                                  version, timeout)

        uri_base = "{}/{}".format(base, version)

        self.voice = Voice(uri_base, self.auth, self.timeout)
        self.phone_numbers = PhoneNumbers(uri_base, self.auth, self.timeout)



import time
import unittest

from twilio import jwt
from twilio.task_router import TaskRouterTaskQueueCapability


class TaskRouterTaskQueueCapabilityTest(unittest.TestCase):

    def setUp(self):
        self.account_sid = "AC123"
        self.auth_token = "foobar"
        self.workspace_sid = "WS456"
        self.taskqueue_sid = "WQ789"
        self.capability = TaskRouterTaskQueueCapability(self.account_sid, self.auth_token, self.workspace_sid, self.taskqueue_sid)

    def test_generate_token(self):

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.assertEqual(decoded["iss"], self.account_sid)
        self.assertEqual(decoded["account_sid"], self.account_sid)
        self.assertEqual(decoded["workspace_sid"], self.workspace_sid)
        self.assertEqual(decoded["taskqueue_sid"], self.taskqueue_sid)
        self.assertEqual(decoded["channel"], self.taskqueue_sid)
        self.assertEqual(decoded["version"], "v1")
        self.assertEqual(decoded["friendly_name"], self.taskqueue_sid)

    def test_generate_token_with_default_ttl(self):
        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.assertEqual(int(time.time()) + 3600, decoded["exp"])

    def test_generate_token_with_custom_ttl(self):
        ttl = 10000

        token = self.capability.generate_token(ttl)
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        self.assertEqual(int(time.time()) + 10000, decoded["exp"])

    def test_default(self):
        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        policies = decoded['policies']
        self.assertEqual(len(policies), 3)

        # websocket GET
        get_policy = policies[0]
        self.assertEqual("https://event-bridge.twilio.com/v1/wschannels/AC123/WQ789", get_policy['url'])
        self.assertEqual("GET", get_policy['method'])
        self.assertTrue(get_policy['allow'])
        self.assertEqual({}, get_policy['query_filter'])
        self.assertEqual({}, get_policy['post_filter'])

        # websocket POST
        post_policy = policies[1]
        self.assertEqual("https://event-bridge.twilio.com/v1/wschannels/AC123/WQ789", post_policy['url'])
        self.assertEqual("POST", post_policy['method'])
        self.assertTrue(post_policy['allow'])
        self.assertEqual({}, post_policy['query_filter'])
        self.assertEqual({}, post_policy['post_filter'])

        # fetch GET
        fetch_policy = policies[2]
        self.assertEqual("https://taskrouter.twilio.com/v1/Workspaces/WS456/TaskQueues/WQ789", fetch_policy['url'])
        self.assertEqual("GET", fetch_policy['method'])
        self.assertTrue(fetch_policy['allow'])
        self.assertEqual({}, fetch_policy['query_filter'])
        self.assertEqual({}, fetch_policy['post_filter'])

    def test_allow_fetch_subresources(self):
        self.capability.allow_fetch_subresources()

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        policies = decoded['policies']
        self.assertEqual(len(policies), 4)

        # confirm the additional policy generated with allow_fetch_subresources()

        policy = policies[3]

        self.assertEqual(policy['url'], "https://taskrouter.twilio.com/v1/Workspaces/WS456/TaskQueues/WQ789/**")
        self.assertEqual(policy['method'], "GET")
        self.assertTrue(policy['allow'])
        self.assertEqual({}, policy['query_filter'])
        self.assertEqual({}, policy['post_filter'])

    def test_allow_updates_subresources(self):
        self.capability.allow_updates_subresources()

        token = self.capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, self.auth_token)
        self.assertNotEqual(None, decoded)

        policies = decoded['policies']
        self.assertEqual(len(policies), 4)

        # confirm the additional policy generated with allow_updates_subresources()

        policy = policies[3]

        self.assertEqual(policy['url'], "https://taskrouter.twilio.com/v1/Workspaces/WS456/TaskQueues/WQ789/**")
        self.assertEqual(policy['method'], "POST")
        self.assertTrue(policy['allow'])
        self.assertEqual({}, policy['query_filter'])
        self.assertEqual({}, policy['post_filter'])

if __name__ == "__main__":
    unittest.main()



from datetime import date
from mock import patch, Mock
from nose.tools import assert_true
from twilio.rest.resources import Calls, Call
from tests.tools import create_mock_json

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")
CALL_SID = "CA47e13748ed59a5733d2c1c1c69a83a28"

list_resource = Calls(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_call(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Calls" % (BASE_URI)
    list_resource.create("TO", "FROM", "url", record=True, application_sid='APPSID')
    exp_params = {
        'To': "TO",
        'From': "FROM",
        'Url': "url",
        'Record': "true",
        'ApplicationSid': 'APPSID',
        }

    mock.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_call_status_events(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Calls" % (BASE_URI)
    list_resource.create("TO", "FROM", "url",
                         status_callback="http://example.com",
                         status_events=['initiated', 'completed'])
    exp_params = {
        'To': "TO",
        'From': "FROM",
        'Url': "url",
        'StatusCallbackEvent': ['initiated', 'completed'],
        'StatusCallback': 'http://example.com',
        }

    mock.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_call_status_events_none(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Calls" % (BASE_URI)
    list_resource.create("TO", "FROM", "url",
                         status_callback="http://example.com")
    exp_params = {
        'To': "TO",
        'From': "FROM",
        'Url': "url",
        'StatusCallback': 'http://example.com',
        }

    mock.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_paging(mock):
    resp = create_mock_json("tests/resources/calls_list.json")
    mock.return_value = resp

    uri = "%s/Calls" % (BASE_URI)
    list_resource.list(started_before=date(2010, 12, 5))
    exp_params = {'StartTime<': '2010-12-05'}

    mock.assert_called_with("GET", uri, params=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_paging_iter(mock):
    resp = create_mock_json("tests/resources/calls_list.json")
    mock.return_value = resp

    uri = "%s/Calls" % (BASE_URI)
    next(list_resource.iter(started_before=date(2010, 12, 5)))
    exp_params = {'StartTime<': '2010-12-05'}

    mock.assert_called_with("GET", uri, params=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    mock.return_value = resp

    uri = "%s/Calls/%s" % (BASE_URI, CALL_SID)
    list_resource.get(CALL_SID)

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_hangup(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    resp.status_code = 204
    mock.return_value = resp

    uri = "%s/Calls/%s" % (BASE_URI, CALL_SID)
    r = list_resource.hangup(CALL_SID)
    exp_data = {"Status": "completed"}

    mock.assert_called_with("POST", uri, data=exp_data, auth=AUTH,
                            use_json_extension=True)
    assert_true(r)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_cancel(mock):
    resp = create_mock_json("tests/resources/calls_instance.json")
    resp.status_code = 204
    mock.return_value = resp

    uri = "%s/Calls/%s" % (BASE_URI, CALL_SID)
    r = list_resource.cancel(CALL_SID)
    exp_data = {"Status": "canceled"}

    mock.assert_called_with("POST", uri, data=exp_data, auth=AUTH,
                            use_json_extension=True)
    assert_true(r)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete(req):
    """ Deleting a call should work """
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    req.return_value = resp, {}

    app = Call(list_resource, "CA123")
    app.delete()
    uri = "https://api.twilio.com/2010-04-01/Accounts/AC123/Calls/CA123"
    req.assert_called_with("DELETE", uri)



from twilio.rest.resources.base import InstanceResource, ListResource


class Key(InstanceResource):
    """
    A key resource.
    See https://www.twilio.com/docs/api/rest-keys

    .. attribute:: sid

        The unique ID for this key.

    .. attribute:: friendly_name

        A human-readable description of this key.

    .. attribute:: secret

        This key's secret.

    .. attribute:: date_created

        The date this key was created, given as UTC in ISO 8601 format.

    .. attribute:: date_updated

        The date this singing key was last updated, given as UTC in ISO 8601
        format.
    """

    def update(self, **kwargs):
        """
        Update this key
        """
        return self.parent.update(self.name, **kwargs)

    def delete(self):
        """
        Delete this key
        """
        return self.parent.delete(self.name)


class Keys(ListResource):
    name = "Keys"
    key = "keys"
    instance = Key

    def create(self, **kwargs):
        """
        Create a :class:`Key` with any of these optional parameters.

        :param friendly_name: A human readable description of the signing key.
        """
        return self.create_instance(kwargs)

    def update(self, sid, **kwargs):
        """
        Update a :class:`Key` with the given parameters.

        All the parameters are describe above in :meth:`create`
        """
        return self.update_instance(sid, kwargs)

    def delete(self, sid):
        """
        Delete a :class:`Key`
        """
        return self.delete_instance(sid)

    def list(self, **kwargs):
        """
        Returns a page of :class:`Key` resources as a list
        """
        return self.get_instances(kwargs)



from datetime import date
import unittest

from mock import patch
from six import u

from twilio.rest.resources import Messages

DEFAULT = {
    'From': None,
    'DateSent<': None,
    'DateSent>': None,
    'DateSent': None,
}


class MessageTest(unittest.TestCase):

    def setUp(self):
        self.resource = Messages("foo", ("sid", "token"))
        self.params = DEFAULT.copy()

    def test_list_on(self):
        with patch.object(self.resource, 'get_instances') as mock:
            self.resource.list(date_sent=date(2011, 1, 1))
            self.params['DateSent'] = "2011-01-01"
            mock.assert_called_with(self.params)

    def test_list_after(self):
        with patch.object(self.resource, 'get_instances') as mock:
            self.resource.list(after=date(2011, 1, 1))
            self.params['DateSent>'] = "2011-01-01"
            mock.assert_called_with(self.params)

    def test_list_before(self):
        with patch.object(self.resource, 'get_instances') as mock:
            self.resource.list(before=date(2011, 1, 1))
            self.params['DateSent<'] = "2011-01-01"
            mock.assert_called_with(self.params)

    def test_create(self):
        with patch.object(self.resource, 'create_instance') as mock:
            self.resource.create(
                from_='+14155551234',
                to='+14155556789',
                body=u('ahoy hoy'),
            )
            mock.assert_called_with(
                {
                    'from': '+14155551234',
                    'to': '+14155556789',
                    'body': u('ahoy hoy'),
                },
            )

    def test_delete(self):
        with patch.object(self.resource, 'delete_instance') as mock:
            self.resource.delete('MM123')
            mock.assert_called_with('MM123')

    def test_redact(self):
        with patch.object(self.resource, 'update_instance') as mock:
            self.resource.redact('MM123')
            mock.assert_called_with('MM123', {'Body': ''})



from . import InstanceResource, ListResource
from .util import normalize_dates, parse_date


class Media(InstanceResource):
    """ Represents media associated with a :class:`Message`.

    .. attribute:: sid

        A 34 character string that uniquely identifies this resource.

    .. attribute:: account_sid

        The unique id of the Account responsible for this media.

    .. attribute:: date_created

        The date that this resource was created, given in RFC 2822 format.

    .. attribute:: date_updated

        The date that this resource was last updated, given in RFC 2822 format.

    .. attribute:: parent_sid

        The MessageSid of the message that created the media.

    .. attribute:: content_type

        The default content-type of the media, for example image/jpeg,
        image/png, or image/gif.

    .. attribute:: uri

        The URI for this resource, relative to https://api.twilio.com

    """

    def delete(self):
        """
        Delete this media.
        """
        return self.parent.delete_instance(self.name)


class MediaList(ListResource):
    name = "Media"
    key = "media_list"
    instance = Media

    def __call__(self, message_sid):
        # `Media` is a word of ambiguous plurality. This causes issues.
        # To match the rest of the library:
        # `client.media` needs to return a new MediaList.
        # `client.media('message_sid')` needs to return a MediaList
        # for a given message.

        base_uri = "%s/Messages/%s" % (self.base_uri, message_sid)
        return MediaList(base_uri, self.auth, self.timeout)

    def __init__(self, *args, **kwargs):
        super(MediaList, self).__init__(*args, **kwargs)

    @normalize_dates
    def list(self, before=None, after=None, date_created=None, **kw):
        """
        Returns a page of :class:`Media` resources as a list. For
        paging information see :class:`ListResource`.

        :param date after: Only list media created after this date.
        :param date before: Only list media created before this date.
        :param date date_created: Only list media created on this date.
        :param sid message_sid: Only list media created by the given MessageSid
        """
        kw["DateCreated<"] = before
        kw["DateCreated>"] = after
        kw["DateCreated"] = parse_date(date_created)
        return self.get_instances(kw)

    def delete(self, sid):
        """
        Delete a :class:`Media`.

        :param sid: String identifier for a Media resource
        """
        return self.delete_instance(sid)



from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Participant(NextGenInstanceResource):
    """A participant in a Conversation.

    .. attribute:: sid

    .. attribute:: conversation_sid

    .. attribute:: account_sid

    .. attribute:: status

    .. attribute:: address

    .. attribute:: date_created

    .. attribute:: start_time

    .. attribute:: end_time

    .. attribute:: duration

    .. attribute:: url
    """
    pass


class Participants(NextGenListResource):
    """A list of :class:`Participant` objects."""

    name = "Participants"
    instance = Participant



from twilio.rest.base import TwilioClient
from twilio.rest.resources import (
    UNSET_TIMEOUT,
    Accounts,
    Applications,
    AuthorizedConnectApps,
    CallFeedback,
    CallFeedbackFactory,
    CallerIds,
    Calls,
    Conferences,
    ConnectApps,
    DependentPhoneNumbers,
    Keys,
    MediaList,
    Members,
    Messages,
    Notifications,
    Participants,
    PhoneNumbers,
    Queues,
    Recordings,
    Sandboxes,
    SigningKeys,
    Sip,
    Sms,
    Tokens,
    Transcriptions,
    Usage,
)


class TwilioRestClient(TwilioClient):
    """
    A client for accessing the Twilio REST API

    :param str account: Your Account SID from `your dashboard
        <https://twilio.com/user/account>`_
    :param str token: Your Auth Token from `your dashboard
        <https://twilio.com/user/account>`_
    :param float timeout: The socket and read timeout for requests to Twilio
    """

    def __init__(self, account=None, token=None, base="https://api.twilio.com",
                 version="2010-04-01", timeout=UNSET_TIMEOUT):
        """
        Create a Twilio REST API client.
        """
        super(TwilioRestClient, self).__init__(account, token, base, version,
                                               timeout)

        version_uri = "%s/%s" % (base, version)

        self.accounts = Accounts(version_uri, self.auth, timeout)
        self.applications = Applications(self.account_uri, self.auth, timeout)
        self.authorized_connect_apps = AuthorizedConnectApps(
            self.account_uri,
            self.auth,
            timeout
        )
        self.calls = Calls(self.account_uri, self.auth, timeout)
        self.caller_ids = CallerIds(self.account_uri, self.auth, timeout)
        self.connect_apps = ConnectApps(self.account_uri, self.auth, timeout)
        self.notifications = Notifications(self.account_uri, self.auth,
                                           timeout)
        self.recordings = Recordings(self.account_uri, self.auth, timeout)
        self.transcriptions = Transcriptions(self.account_uri, self.auth,
                                             timeout)
        self.sms = Sms(self.account_uri, self.auth, timeout)
        self.phone_numbers = PhoneNumbers(self.account_uri, self.auth, timeout)
        self.conferences = Conferences(self.account_uri, self.auth, timeout)
        self.queues = Queues(self.account_uri, self.auth, timeout)
        self.sandboxes = Sandboxes(self.account_uri, self.auth, timeout)
        self.usage = Usage(self.account_uri, self.auth, timeout)
        self.messages = Messages(self.account_uri, self.auth, timeout)
        self.media = MediaList(self.account_uri, self.auth, timeout)
        self.sip = Sip(self.account_uri, self.auth, timeout)
        self.signing_keys = SigningKeys(self.account_uri, self.auth, timeout)
        self.tokens = Tokens(self.account_uri, self.auth, timeout)
        self.keys = Keys(self.account_uri, self.auth, timeout)

    def participants(self, conference_sid):
        """
        Return a :class:`~twilio.rest.resources.Participants` instance for the
        :class:`~twilio.rest.resources.Conference` with given conference_sid
        """
        base_uri = "%s/Conferences/%s" % (self.account_uri, conference_sid)
        return Participants(base_uri, self.auth, self.timeout)

    def members(self, queue_sid):
        """
        Return a :class:`Members <twilio.rest.resources.Members>` instance for
        the :class:`Queue <twilio.rest.resources.Queue>` with the
        given queue_sid
        """
        base_uri = "%s/Queues/%s" % (self.account_uri, queue_sid)
        return Members(base_uri, self.auth, self.timeout)

    def feedback(self, call_sid):
        """
        Return a :class:`CallFeedback <twilio.rest.resources.CallFeedback>`
        instance for the :class:`Call <twilio.rest.resources.calls.Call>`
        with the given call_sid
        """
        base_uri = "%s/Calls/%s/Feedback" % (self.account_uri, call_sid)
        call_feedback_list = CallFeedbackFactory(
            base_uri,
            self.auth,
            self.timeout
        )
        return CallFeedback(call_feedback_list)

    def dependent_phone_numbers(self, address_sid):
        """
        Return a :class:`DependentPhoneNumbers
        <twilio.rest.resources.DependentPhoneNumbers>` instance for the
        :class:`Address <twilio.rest.resources.Address>` with the given
        address_sid
        """
        base_uri = "%s/Addresses/%s" % (self.account_uri, address_sid)
        return DependentPhoneNumbers(base_uri, self.auth, self.timeout)



# -*- coding: utf-8 -*-
from mock import patch, Mock
from six import u
from twilio.rest import resources


@patch("httplib2.Http")
@patch("twilio.rest.resources.base.Response")
def test_ascii_encode(resp_mock, mock):
    http = mock.return_value
    http.request.return_value = (Mock(), Mock())

    data = {
        "body": "HeyHey".encode('utf-8')
    }

    resources.make_request("GET", "http://www.example.com", data=data)

    http.request.assert_called_with("http://www.example.com", "GET",
                                    headers=None, body="body=HeyHey")


@patch("httplib2.Http")
@patch("twilio.rest.resources.base.Response")
def test_ascii(resp_mock, mock):
    http = mock.return_value
    http.request.return_value = (Mock(), Mock())

    data = {
        "body": "HeyHey"
    }

    resources.make_request("GET", "http://www.example.com", data=data)

    http.request.assert_called_with("http://www.example.com", "GET",
                                    headers=None, body="body=HeyHey")


@patch("httplib2.Http")
@patch("twilio.rest.resources.base.Response")
def test_double_encoding(resp_mock, mock):
    http = mock.return_value
    http.request.return_value = (Mock(), Mock())

    body = u('Chlo\xe9\xf1')

    data = {
        "body": body.encode('utf-8'),
    }

    resources.make_request("GET", "http://www.example.com", data=data)

    http.request.assert_called_with("http://www.example.com", "GET",
                                    headers=None, body="body=Chlo%C3%A9%C3%B1")


@patch("httplib2.Http")
@patch("twilio.rest.resources.base.Response")
def test_paging(resp_mock, mock):
    http = mock.return_value
    http.request.return_value = (Mock(), Mock())

    data = {
        "body": u('Chlo\xe9\xf1'),
    }

    resources.make_request("GET", "http://www.example.com", data=data)

    http.request.assert_called_with("http://www.example.com", "GET",
                                    headers=None, body="body=Chlo%C3%A9%C3%B1")


@patch("httplib2.Http")
@patch("twilio.rest.resources.base.Response")
def test_unicode_sequence_form_value(resp_mock, mock):
    http = mock.return_value
    http.request.return_value = (Mock(), Mock())

    data = {
        "body": [u('\xe5'), u('\xe7')],
    }

    resources.make_request("POST", "http://www.example.com", data=data)

    http.request.assert_called_with(
        "http://www.example.com",
        "POST",
        headers=None,
        body="body=%C3%A5&body=%C3%A7",
    )



from .workflow_ruletarget import WorkflowRuleTarget


class WorkflowRule:

    """
    WorkflowRule represents the top level filter
    which contains a 1 or more targets

    ..attribute::expression

       The expression at the top level filter

    ..attribute::targets

       The list of targets under the filter

    ..attribute::friendlyName

       The name of the filter
    """

    def __init__(self, expression, targets, friendly_name):

        self.expression = expression
        self.targets = targets
        self.friendly_name = friendly_name

    def __repr__(self):
        return str({
            'expression': self.expression,
            'friendly_name': self.friendly_name,
            'target': self.target,
        })



from __future__ import with_statement
import sys
from setuptools import setup, find_packages

__version__ = None
with open('twilio/version.py') as f:
    exec(f.read())

# To install the twilio-python library, open a Terminal shell, then run this
# file by typing:
#
# python setup.py install
#
# You need to have the setuptools module installed. Try reading the setuptools
# documentation: http://pypi.python.org/pypi/setuptools
REQUIRES = ["httplib2 >= 0.7", "six", "pytz"]

if sys.version_info < (2, 6):
    REQUIRES.append('simplejson')
if sys.version_info >= (3,0):
    REQUIRES.append('pysocks')

setup(
    name = "twilio",
    version = __version__,
    description = "Twilio API client and TwiML generator",
    author = "Twilio",
    author_email = "help@twilio.com",
    url = "https://github.com/twilio/twilio-python/",
    keywords = ["twilio","twiml"],
    install_requires = REQUIRES,
    # bdist conditional requirements support
    extras_require={
        ':python_version=="3.2"': ['pysocks'],
        ':python_version=="3.3"': ['pysocks'],
        ':python_version=="3.4"': ['pysocks'],
    },
    packages = find_packages(),
    include_package_data=True,
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Telephony",
        ],
    long_description = """\
    Python Twilio Helper Library
    ----------------------------

    DESCRIPTION
    The Twilio REST SDK simplifies the process of making calls using the Twilio REST API.
    The Twilio REST API lets to you initiate outgoing calls, list previous calls,
    and much more.  See https://www.github.com/twilio/twilio-python for more information.

     LICENSE The Twilio Python Helper Library is distributed under the MIT
    License """ )



from . import InstanceResource, ListResource
from .util import UNSET_TIMEOUT


class UsageTrigger(InstanceResource):
    """ A usage trigger resource """

    def update(self, **kwargs):
        """
        Update this usage trigger
        """
        return self.parent.update(self.name, **kwargs)

    def delete(self):
        """
        Delete this usage trigger
        """
        return self.parent.delete(self.name)


class UsageTriggers(ListResource):
    name = "Usage/Triggers"
    instance = UsageTrigger
    key = "usage_triggers"

    def list(self, **kwargs):
        """
        Returns a page of :class:`UsageTrigger` resources as a list. For paging
        information see :class:`ListResource`

        :param recurring: Only show UsageTriggers that count over this
                          interval. One of daily, monthly, or yearly. To
                          retrieve non-recurring triggers, leave this empty or
                          use alltime.
        :param usage_category: Only show UsageTriggers that watch this usage
                               category. Must be one of the supported usage
                               categories.
        :trigger_by: Only show UsageTriggers that trigger by this field in the
                     UsageRecord. Must be one of: count, usage, or price as
                     described in the UsageRecords documentation.
        """
        return self.get_instances(kwargs)

    def create(self, **kwargs):
        """
        Create an :class:`UsageTrigger` with any of these optional parameters.

        :param friendly_name: A human readable description of the UsageTrigger,
                              with maximum length 64 characters.
        :param usage_category: The trigger will watch this usage category.
        :param trigger_value: The trigger will fire when usage reaches this
                              value.
        :param callback_url: Twilio will make a request to this url when the
                             trigger fires.
        :param trigger_by: The field in the UsageRecord that will fire the
                           trigger. One of count, usage, or price as described
                           in the UsageRecords documentation. The default is
                           usage.
        :param recurring: The interval the trigger will count over. Must be one
                          of: daily, monthly, or yearly. Omit this to create a
                          non-recurring trigger.
        :param callback_method: Twilio will use this HTTP method when making a
                                request to the CallbackUrl. GET or POST. The
                                default is POST.
        """
        return self.create_instance(kwargs)

    def update(self, sid, **kwargs):
        """ Update the UsageTrigger with the given sid to have the kwargs """
        return self.update_instance(sid, kwargs)

    def delete(self, sid):
        """
        Delete a :class:`UsageTrigger`
        """
        return self.delete_instance(sid)


class UsageRecord(InstanceResource):
    """ A usage record resource """

    def load(self, entries):
        uri = entries.get('uri')
        super(UsageRecord, self).load(entries)
        self.__dict__.update({'uri': uri})

    @property
    def uri(self):
        return self.__dict__.get('uri')


class BaseUsageRecords(ListResource):
    name = "Usage/Records"
    instance = UsageRecord
    key = "usage_records"

    def list(self, **kwargs):
        """
        Returns a page of :class:`UsageRecord` resources as a list. For paging
        information see :class:`ListResource`

        :param category: Only include usage of this usage category.
        :param start_date: Only include usage that has occurred on or after
                           this date. Format is YYYY-MM-DD. All dates are in
                           GMT.
        :param end_date: Only include usage that has occurred on or before this
                         date. Format is YYYY-MM-DD. All dates are in GMT.
        """
        return self.get_instances(kwargs)

    def get(self, *args, **kwargs):
        raise AttributeError('Unsupported in UsageRecords')

    def load_instance(self, data):
        instance = self.instance(self, "Resource")
        instance.load(data)
        instance.load_subresources()
        return instance


class UsageRecords(BaseUsageRecords):

    def __init__(self, base_uri, auth, timeout=UNSET_TIMEOUT):
        super(UsageRecords, self).__init__(base_uri, auth, timeout)
        self.daily = UsageRecordsDaily(base_uri, auth, timeout)
        self.monthly = UsageRecordsMonthly(base_uri, auth, timeout)
        self.yearly = UsageRecordsYearly(base_uri, auth, timeout)
        self.today = UsageRecordsToday(base_uri, auth, timeout)
        self.yesterday = UsageRecordsYesterday(base_uri, auth, timeout)
        self.this_month = UsageRecordsThisMonth(base_uri, auth, timeout)
        self.last_month = UsageRecordsLastMonth(base_uri, auth, timeout)


class UsageRecordsDaily(BaseUsageRecords):
    name = "Usage/Records/Daily"


class UsageRecordsMonthly(BaseUsageRecords):
    name = "Usage/Records/Monthly"


class UsageRecordsYearly(BaseUsageRecords):
    name = "Usage/Records/Yearly"


class UsageRecordsToday(BaseUsageRecords):
    name = "Usage/Records/Today"


class UsageRecordsYesterday(BaseUsageRecords):
    name = "Usage/Records/Yesterday"


class UsageRecordsThisMonth(BaseUsageRecords):
    name = "Usage/Records/ThisMonth"


class UsageRecordsLastMonth(BaseUsageRecords):
    name = "Usage/Records/LastMonth"

UsageRecord.subresources = [
    UsageRecordsDaily,
    UsageRecordsMonthly,
    UsageRecordsYearly,
    UsageRecordsToday,
    UsageRecordsYesterday,
    UsageRecordsThisMonth,
    UsageRecordsLastMonth
]


class Usage(object):
    """
    Holds all the specific Usage list resources
    """

    def __init__(self, base_uri, auth, timeout=UNSET_TIMEOUT):
        self.records = UsageRecords(base_uri, auth, timeout=timeout)
        self.triggers = UsageTriggers(base_uri, auth, timeout=timeout)
        self.timeout = timeout



import unittest
from nose.tools import assert_equal

from mock import patch
from tests.tools import create_mock_json

from twilio.rest.resources.pricing.phone_numbers import PhoneNumberCountries


AUTH = ("AC123", "token")
BASE_URI = "https://pricing.twilio.com/v1"


class NumbersTest(unittest.TestCase):

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_number_countries(self, request):
        resp = create_mock_json('tests/resources/pricing/phone_number_country_list.json')
        resp.status_code = 200
        request.return_value = resp

        countries = PhoneNumberCountries(BASE_URI + "/PhoneNumbers", AUTH)
        result = countries.list()

        assert_equal(result[0].iso_country, "AC")
        assert_equal(len(result), 3)

        request.assert_called_with(
            "GET",
            "{0}/PhoneNumbers/Countries".format(BASE_URI),
            auth=AUTH,
            use_json_extension=False,
        )

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_number_country(self, request):
        resp = create_mock_json('tests/resources/pricing/phone_number_country_instance.json')
        resp.status_code = 200
        request.return_value = resp

        countries = PhoneNumberCountries(BASE_URI + "/PhoneNumbers", AUTH)
        country = countries.get('EE')

        assert_equal(country.country, "Estonia")
        assert_equal(
            country.phone_number_prices,
            [
                {
                    'type': 'mobile',
                    'base_price': 3.00,
                    'current_price': 3.00,
                },
                {
                    'type': 'national',
                    'base_price': 1.00,
                    'current_price': 1.00,
                }
            ],
        )

        request.assert_called_with(
            "GET",
            "{0}/PhoneNumbers/Countries/EE".format(BASE_URI),
            auth=AUTH,
            use_json_extension=False,
        )



import time
import unittest

from twilio import jwt

from twilio.task_router import TaskRouterCapability


class TaskRouterCapabilityTest(unittest.TestCase):

    def setUp(self):
        self.account_sid = "AC123"
        self.auth_token = "foobar"
        self.workspace_sid = "WS456"
        self.worker_sid = "WK789"
        self.cap = TaskRouterCapability(self.account_sid, self.auth_token,
                                        self.workspace_sid, self.worker_sid)

    def test_generate_token(self):
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        del decoded['exp']
        del decoded['policies']

        expected = {
            'iss': self.account_sid,
            'account_sid': self.account_sid,
            'workspace_sid': self.workspace_sid,
            'worker_sid': self.worker_sid,
            'channel': self.worker_sid,
            'version': 'v1',
            'friendly_name': self.worker_sid,
        }
        self.assertEqual(expected, decoded)

    def test_generate_token_default_ttl(self):
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        self.assertEqual(int(time.time()) + 3600, decoded['exp'])

    def test_generate_token_custom_ttl(self):
        token = self.cap.generate_token(10000)
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        self.assertEqual(int(time.time()) + 10000, decoded['exp'])

    def test_defaults(self):
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        websocket_url = (
            'https://event-bridge.twilio.com/v1/wschannels/{0}/{1}'.format(self.account_sid, self.worker_sid)
        )
        expected = [
            {
                'url': 'https://taskrouter.twilio.com/v1/Workspaces/WS456/Activities',
                'method': 'GET',
                'allow': True,
                'query_filter': {},
                'post_filter': {},
            },
            {
                'url': 'https://taskrouter.twilio.com/v1/Workspaces/{0}/Tasks/**'.format(self.workspace_sid),
                'method': 'GET',
                'allow': True,
                'query_filter': {},
                'post_filter': {},
            },
            {
                'url': websocket_url,
                'method': 'GET',
                'allow': True,
                'query_filter': {},
                'post_filter': {},
            },
            {
                'url': websocket_url,
                'method': 'POST',
                'allow': True,
                'query_filter': {},
                'post_filter': {},
            },
            {
                'url': 'https://taskrouter.twilio.com/v1/Workspaces/{0}/Workers/{1}'.format(self.workspace_sid, self.worker_sid),
                'method': 'GET',
                'allow': True,
                'query_filter': {},
                'post_filter': {},
            }
        ]
        self.assertEqual(expected, decoded['policies'])

    def test_allow_worker_activity_updates(self):
        self.cap.allow_worker_activity_updates()
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        url = 'https://taskrouter.twilio.com/v1/Workspaces/{0}/Workers/{1}'.format(
            self.workspace_sid,
            self.worker_sid,
        )

        expected = {
            'url': url,
            'method': 'POST',
            'allow': True,
            'query_filter': {},
            'post_filter': {'ActivitySid': {'required': True}},
        }
        self.assertEqual(expected, decoded['policies'][-1])

    def test_allow_worker_fetch_attributes(self):
        self.cap.allow_worker_fetch_attributes()
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        url = 'https://taskrouter.twilio.com/v1/Workspaces/{0}/Workers/{1}'.format(
            self.workspace_sid,
            self.worker_sid,
        )

        expected = {
            'url': url,
            'method': 'GET',
            'allow': True,
            'query_filter': {},
            'post_filter': {},
        }

        self.assertEqual(expected, decoded['policies'][-1])

    def test_allow_task_reservation_updates(self):
        self.cap.allow_task_reservation_updates()
        token = self.cap.generate_token()
        decoded = jwt.decode(token, self.auth_token)

        self.assertTrue(decoded is not None)
        url = 'https://taskrouter.twilio.com/v1/Workspaces/{0}/Tasks/**'.format(
            self.workspace_sid,
        )

        expected = {
            'url': url,
            'method': 'POST',
            'allow': True,
            'query_filter': {},
            'post_filter': {},
        }
        self.assertEqual(expected, decoded['policies'][-1])



import unittest

from mock import patch, Mock

from tests.tools import create_mock_json
from twilio.rest.resources.task_router.task_queues import TaskQueues, TaskQueue


AUTH = ("AC123", "token")
BASE_URI = "https://taskrouter.twilio.com/v1/Accounts/AC123/Workspaces/WSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
TASK_QUEUE_SID = "WQaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
TIMEOUT = 30


class TaskQueueTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create(self, request):
        resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
        resp.status_code = 201
        request.return_value = resp

        task_queues = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        task_queues.create("Test TaskQueue", "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
        exp_params = {
            'FriendlyName': "Test TaskQueue",
            'AssignmentActivitySid': "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'ReservationActivitySid': 'WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        }

        request.assert_called_with("POST",
                                   "{0}/TaskQueues".format(BASE_URI),
                                   data=exp_params, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_instance(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/TaskQueues/{1}".format(BASE_URI, TASK_QUEUE_SID)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        task_queue = TaskQueue(list_resource, TASK_QUEUE_SID)
        task_queue.delete()
        request.assert_called_with("DELETE", uri, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_delete_list(self, request):
        resp = Mock()
        resp.content = ""
        resp.status_code = 204
        request.return_value = resp

        uri = "{0}/TaskQueues/{1}".format(BASE_URI, TASK_QUEUE_SID)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        list_resource.delete(TASK_QUEUE_SID)
        request.assert_called_with("DELETE", uri, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/TaskQueues/{1}".format(BASE_URI, TASK_QUEUE_SID)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        list_resource.get(TASK_QUEUE_SID)
        request.assert_called_with("GET", uri, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/task_router/task_queues_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/TaskQueues".format(BASE_URI)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_instance(self, request):
        resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/TaskQueues/{1}".format(BASE_URI, TASK_QUEUE_SID)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        task_queue = TaskQueue(list_resource, TASK_QUEUE_SID)
        task_queue.update(friendly_name='Test TaskQueue', assignment_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
                          reservation_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        exp_params = {
            'FriendlyName': "Test TaskQueue",
            'AssignmentActivitySid': "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'ReservationActivitySid': 'WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        }

        request.assert_called_with("POST", uri, data=exp_params,
                                   auth=AUTH, timeout=TIMEOUT,
                                   use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_update_list(self, request):
        resp = create_mock_json('tests/resources/task_router/task_queues_instance.json')
        resp.status_code = 201
        request.return_value = resp

        uri = "{0}/TaskQueues/{1}".format(BASE_URI, TASK_QUEUE_SID)
        list_resource = TaskQueues(BASE_URI, AUTH, TIMEOUT)
        list_resource.update(TASK_QUEUE_SID, friendly_name='Test TaskQueue',
                             assignment_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
                             reservation_activity_sid='WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        exp_params = {
            'FriendlyName': "Test TaskQueue",
            'AssignmentActivitySid': "WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            'ReservationActivitySid': 'WAaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        }

        request.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                                   timeout=TIMEOUT, use_json_extension=False)



import unittest

from mock import patch

from tests.tools import create_mock_json
from twilio.rest.resources.monitor.events import Events


AUTH = ("AC123", "token")
BASE_URI = "https://monitor.twilio.com/v1"
EVENT_SID = "AEaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


class EventTest(unittest.TestCase):
    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get(self, request):
        resp = create_mock_json('tests/resources/monitor/events_instance.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Events/{1}".format(BASE_URI, EVENT_SID)
        list_resource = Events(BASE_URI, AUTH)
        list_resource.get(EVENT_SID)
        request.assert_called_with("GET", uri, auth=AUTH, use_json_extension=False)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_list(self, request):
        resp = create_mock_json('tests/resources/monitor/events_list.json')
        resp.status_code = 200
        request.return_value = resp

        uri = "{0}/Events".format(BASE_URI)
        list_resource = Events(BASE_URI, AUTH)
        list_resource.list()
        request.assert_called_with("GET", uri, params={}, auth=AUTH, use_json_extension=False)



import time
import jwt

ALL = '*'

# HTTP Actions
HTTP_DELETE = 'DELETE'
HTTP_GET = 'GET'
HTTP_POST = 'POST'
HTTP_PUT = 'PUT'

# Client Actions
CLIENT_LISTEN = 'listen'
CLIENT_INVITE = 'invite'


class AccessToken(object):
    def __init__(self, signing_key_sid, account_sid, secret, ttl=3600):
        self.signing_key_sid = signing_key_sid
        self.account_sid = account_sid
        self.secret = secret
        self.ttl = ttl
        self.grants = []

    def add_grant(self, resource, actions=ALL):
        if not isinstance(actions, list):
            actions = [actions]

        self.grants.append({
            'res': resource,
            'act': actions,
        })
        return self

    def add_rest_grant(self, uri, actions=ALL):
        resource = 'https://api.twilio.com/2010-04-01/Accounts/{0}/{1}'.format(
            self.account_sid,
            uri.lstrip('/'),
        )
        return self.add_grant(resource, actions)

    def add_endpoint_grant(self, endpoint, actions=None):
        actions = actions or [CLIENT_LISTEN, CLIENT_INVITE]
        resource = 'sip:{0}@{1}.endpoint.twilio.com'.format(
            endpoint,
            self.account_sid
        )
        return self.add_grant(resource, actions)

    def enable_nts(self):
        return self.add_rest_grant('/Tokens.json', HTTP_POST)

    def to_jwt(self):
        now = int(time.time())
        headers = {
            "cty": "twilio-sat;v=1"
        }
        payload = {
            "jti": '{0}-{1}'.format(self.signing_key_sid, now),
            "iss": self.signing_key_sid,
            "sub": self.account_sid,
            "nbf": now,
            "exp": now + self.ttl,
            "grants": self.grants
        }

        return jwt.encode(payload, self.secret, headers=headers)

    def __str__(self):
        return self.to_jwt().decode('utf-8')



import unittest

from nose.tools import assert_equal
from mock import Mock, patch, ANY
from tests.tools import create_mock_json
from twilio.rest.resources import Call, Calls

AUTH = ('foo', 'bar')


class CallFeedbackTest(unittest.TestCase):

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get_call_feedback(self, request):
        resp = create_mock_json('tests/resources/call_feedback.json')
        request.return_value = resp

        mock = Mock()
        mock.uri = '/base'
        call = Call(mock, 'CA123')
        call.load_subresources()
        feedback = call.feedback.get()
        assert_equal(5, feedback.quality_score, 5)
        assert_equal(['imperfect-audio', 'post-dial-delay'], feedback.issues)

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create_call_feedback(self, request):
        resp = create_mock_json('tests/resources/call_feedback.json')
        resp.status_code = 201
        request.return_value = resp

        mock = Mock()
        mock.uri = '/base'
        mock.auth = AUTH
        call = Call(mock, 'CA123')
        call.load_subresources()
        feedback = call.feedback.create(
            quality_score=5,
            issues=['imperfect-audio', 'post-dial-delay'],
        )

        exp_data = {
            'QualityScore': 5,
            'Issues': ['imperfect-audio', 'post-dial-delay'],
        }
        assert_equal(5, feedback.quality_score, 5)
        assert_equal(['imperfect-audio', 'post-dial-delay'], feedback.issues)
        request.assert_called_with(
            "POST", "/base/CA123/Feedback",
            data=exp_data, auth=AUTH,
            timeout=ANY, use_json_extension=True,
        )

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_create_call_feedback_one_request(self, request):
        resp = create_mock_json('tests/resources/call_feedback.json')
        resp.status_code = 201
        request.return_value = resp

        base_uri = 'https://api.twilio.com/2010-04-01/Accounts/AC123'
        account_sid = 'AC123'
        auth = (account_sid, "token")

        calls = Calls(base_uri, auth)
        uri = "%s/Calls/CA123/Feedback" % base_uri
        feedback = calls.feedback(
            'CA123',
            quality_score=5,
            issue=['imperfect-audio', 'post-dial-delay']
        )

        exp_data = {
            'QualityScore': 5,
            'Issue': ['imperfect-audio', 'post-dial-delay'],
        }

        assert_equal(['imperfect-audio', 'post-dial-delay'], feedback.issues)
        request.assert_called_with(
            "POST", uri,
            data=exp_data, auth=auth,
            use_json_extension=True,
        )


class CallFeedbackSummaryTest(unittest.TestCase):

    @patch('twilio.rest.resources.base.make_twilio_request')
    def test_get_call_feedback_summary(self, request):
        resp = create_mock_json('tests/resources/call_feedback_summary.json')
        request.return_value = resp

        base_uri = 'https://api.twilio.com/2010-04-01/Accounts/AC123'
        account_sid = 'AC123'
        auth = (account_sid, "token")

        calls = Calls(base_uri, auth)
        uri = "%s/Calls/Summary" % base_uri
        feedback = calls.summary.get()
        assert_equal(10200, feedback.call_count)
        assert_equal(729, feedback.call_feedback_count)

        request.assert_called_with('GET', uri, params={}, auth=auth,
                                   use_json_extension=True)



from .imports import (
    httplib2,
    socks,
    PROXY_TYPE_HTTP,
    PROXY_TYPE_SOCKS4,
    PROXY_TYPE_SOCKS5
)


class Connection(object):
    '''Class for setting proxy configuration to be used for REST calls.'''
    _proxy_info = None

    @classmethod
    def proxy_info(cls):
        '''Returns the currently-set proxy information
        as an httplib2.ProxyInfo object.
        '''
        return cls._proxy_info

    @classmethod
    def set_proxy_info(cls, proxy_host, proxy_port,
                       proxy_type=PROXY_TYPE_HTTP, proxy_rdns=None,
                       proxy_user=None, proxy_pass=None):
        '''Set proxy configuration for future REST API calls.

        :param str proxy_host: Hostname of the proxy to use.
        :param int proxy_port: Port to connect to.
        :param proxy_type: The proxy protocol to use. One of
        PROXY_TYPE_HTTP, PROXY_TYPE_SOCKS4, PROXY_TYPE_SOCKS5.
        Defaults to connection.PROXY_TYPE_HTTP.
        :param bool proxy_rdns: Use the proxy host's DNS resolver.
        :param str proxy_user: Username for the proxy.
        :param str proxy_pass: Password for the proxy.
        '''

        cls._proxy_info = httplib2.ProxyInfo(
            proxy_type,
            proxy_host,
            proxy_port,
            proxy_rdns=proxy_rdns,
            proxy_user=proxy_user,
            proxy_pass=proxy_pass,
        )


_hush_pyflakes = [
    socks,
    PROXY_TYPE_SOCKS4,
    PROXY_TYPE_SOCKS5
]



from datetime import date
from mock import patch
from nose.tools import raises, assert_true
from twilio.rest.resources import Notifications
from tests.tools import create_mock_json

BASE_URI = "https://api.twilio.com/2010-04-01/Accounts/AC123"
ACCOUNT_SID = "AC123"
AUTH = (ACCOUNT_SID, "token")

RE_SID = "RE19e96a31ed59a5733d2c1c1c69a83a28"

list_resource = Notifications(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_paging(mock):
    resp = create_mock_json("tests/resources/notifications_list.json")
    mock.return_value = resp

    uri = "%s/Notifications" % (BASE_URI)
    list_resource.list(before=date(2010, 12, 5))
    exp_params = {'MessageDate<': '2010-12-05'}

    mock.assert_called_with("GET", uri, params=exp_params, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/notifications_instance.json")
    mock.return_value = resp

    uri = "%s/Notifications/%s" % (BASE_URI, RE_SID)
    list_resource.get(RE_SID)

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=True)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get2(mock):
    resp = create_mock_json("tests/resources/notifications_instance.json")
    resp.status_code = 204
    mock.return_value = resp

    uri = "%s/Notifications/%s" % (BASE_URI, RE_SID)
    r = list_resource.delete(RE_SID)

    mock.assert_called_with("DELETE", uri, auth=AUTH,
                            use_json_extension=True)
    assert_true(r)


@raises(AttributeError)
def test_create():
    list_resource.create


@raises(AttributeError)
def test_update():
    list_resource.update



from twilio.exceptions import TwilioException
from twilio.rest.resources import InstanceResource, ListResource


class DependentPhoneNumber(InstanceResource):
    """A purchased phone number that depends on a particular
    :class:`Address`.

    Attributes are the same as :class:`PhoneNumber`.

    DependentPhoneNumbers are a read-only resource and cannot
    be updated or deleted.
    """
    pass


class DependentPhoneNumbers(ListResource):
    """A list of purchased phone numbers that depend on a particular
    :class:`Address`.

    Included numbers are those that require an
    address on file and have no other candidate addresses of the appropriate
    type (local, foreign) associated with the owning account.

    If this list has entries for a given Address, that address cannot be
    deleted until the numbers are released from your account or alternate
    addresses are provided to satisfy the requirements.

    This resource is read-only and cannot be updated or deleted, but will
    reflect the current state of the owning account's addresses (i.e. if
    you add another address that satisfies a number's requirements, it will
    not appear in subsequent requests to this list resource).
    """
    name = "DependentPhoneNumbers"
    key = "dependent_phone_numbers"
    instance = DependentPhoneNumber


class Address(InstanceResource):
    """An Address resource. See https://www.twilio.com/docs/api/rest/address

    .. attribute:: friendly_name

        A human-readable description of this address. Maximum 64 characters.

    .. attribute:: customer_name

        Your or your customer's name or business name.

    .. attribute:: street

        The number and street address where you or your customer are located.

    .. attribute:: city

        The city in which you or your customer are located.

    .. attribute:: region

        The state or region in which you or your customer are located.

    .. attribute:: postal_code

        The postal code in which you or your customer are located.

    .. attribute:: iso_country

        The ISO country code of your or your customer's address.
    """
    subresources = [DependentPhoneNumbers]

    def update(self, **kwargs):
        """Update this phone number instance.

        Parameters are as described in :meth:`Addresses.create`, with
        the exception that `iso_country` cannot be updated on an existing
        Address (create a new one instead).
        """
        return self.parent.update(self.sid, kwargs)


class Addresses(ListResource):
    name = "Addresses"
    key = "addresses"
    instance = Address

    def list(self, customer_name=None, friendly_name=None, iso_country=None):
        kwargs = {
            'customer_name': customer_name,
            'friendly_name': friendly_name,
            'iso_country': iso_country,
        }
        return self.get_instances(kwargs)

    def create(self, customer_name, street, city, region, postal_code,
               iso_country, friendly_name=None):
        """Create an :class:`Address`.

        :param str customer_name: Your customer's name
        :param str street: The number and street of your address
        :param str city: The city of you or your customer's address
        :param str region: The region or state
        :param str postal_code: The postal code of your address
        :param str iso_country: The ISO 3166-1 alpha-2 (two-character)
            country code, e.g. 'US' or 'AU'
        :param str friendly_name: A user-defined name for this address
            (optional; up to 64 characters)
        """
        kwargs = {
            'customer_name': customer_name,
            'street': street,
            'city': city,
            'region': region,
            'postal_code': postal_code,
            'iso_country': iso_country,
        }

        if friendly_name is not None:
            kwargs['friendly_name'] = friendly_name

        return self.create_instance(kwargs)

    def update(self, sid, **kwargs):
        """Update an :class:`Address` with the given parameters.

        Parameters are described above in :meth:`create`, with
        the exception that `iso_country` cannot be updated on
        an existing Address (create a new one instead).
        """
        if 'iso_country' in kwargs:
            raise TwilioException(
                "Cannot update iso_country on an existing Address",
            )

        return self.update_instance(sid, kwargs)

    def delete(self, sid):
        """Delete an :class:`Address`.

        :param str sid: The sid of the Address to delete.
        """
        return self.delete_instance(sid)



from .services import (
    Service,
    Services
)

from .channels import (
    Channel,
    Channels
)

from .members import (
    Member,
    Members
)

from .messages import (
    Message,
    Messages
)

from .roles import (
    Role,
    Roles
)

from .users import (
    User,
    Users
)

from .credentials import (
    Credential,
    Credentials
)



import unittest
import warnings

from twilio import jwt
from twilio.task_router import TaskRouterCapability


class TaskRouterCapabilityTest(unittest.TestCase):

    def check_policy(self, method, url, policy):
        self.assertEqual(url, policy['url'])
        self.assertEqual(method, policy['method'])
        self.assertTrue(policy['allow'])
        self.assertEqual({}, policy['query_filter'])
        self.assertEqual({}, policy['post_filter'])

    def check_decoded(self, decoded, account_sid, workspace_sid, channel_id, channel_sid=None):
        self.assertEqual(decoded["iss"], account_sid)
        self.assertEqual(decoded["account_sid"], account_sid)
        self.assertEqual(decoded["workspace_sid"], workspace_sid)
        self.assertEqual(decoded["channel"], channel_id)
        self.assertEqual(decoded["version"], "v1")
        self.assertEqual(decoded["friendly_name"], channel_id)

        if 'worker_sid' in decoded.keys():
            self.assertEqual(decoded['worker_sid'], channel_sid)
        if 'taskqueue_sid' in decoded.keys():
            self.assertEqual(decoded['taskqueue_sid'], channel_sid)

    def test_workspace_default(self):
        account_sid = "AC123"
        auth_token = "foobar"
        workspace_sid = "WS456"
        channel_id = "WS456"
        capability = TaskRouterCapability(account_sid, auth_token, workspace_sid, channel_id)

        capability.generate_token()

        token = capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, auth_token)
        self.assertNotEqual(None, decoded)

        self.check_decoded(decoded, account_sid, workspace_sid, channel_id)

        policies = decoded['policies']
        self.assertEqual(len(policies), 3)

        for method, url, policy in [
            ('GET', "https://event-bridge.twilio.com/v1/wschannels/AC123/WS456", policies[0]),
            ('POST', "https://event-bridge.twilio.com/v1/wschannels/AC123/WS456", policies[1]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456", policies[2]),
        ]:
            yield self.check_policy, method, url, policy

    def test_worker_default(self):
        account_sid = "AC123"
        auth_token = "foobar"
        workspace_sid = "WS456"
        worker_sid = "WK789"
        capability = TaskRouterCapability(account_sid, auth_token, workspace_sid, worker_sid)

        capability.generate_token()

        token = capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, auth_token)
        self.assertNotEqual(None, decoded)

        self.check_decoded(decoded, account_sid, workspace_sid, worker_sid, worker_sid)

        policies = decoded['policies']
        self.assertEqual(len(policies), 5)

        for method, url, policy in [
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Activities", policies[0]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Tasks/**", policies[1]),
            ('GET', "https://taskrouter.twilio.com/v1/wschannels/AC123/WK789", policies[2]),
            ('POST', "https://event-bridge.twilio.com/v1/wschannels/AC123/WK789", policies[3]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Workers/WK789", policies[4])
        ]:
            yield self.check_policy, method, url, policy

    def test_task_queue_default(self):
        account_sid = "AC123"
        auth_token = "foobar"
        workspace_sid = "WS456"
        taskqueue_sid = "WQ789"
        capability = TaskRouterCapability(account_sid, auth_token, workspace_sid, taskqueue_sid)

        capability.generate_token()

        token = capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, auth_token)
        self.assertNotEqual(None, decoded)

        self.check_decoded(decoded, account_sid, workspace_sid, taskqueue_sid, taskqueue_sid)

        policies = decoded['policies']
        self.assertEqual(len(policies), 3)

        for method, url, policy in [
            ('GET', "https://event-bridge.twilio.com/v1/wschannels/AC123/WQ789", policies[0]),
            ('POST', "https://event-bridge.twilio.com/v1/wschannels/AC123/WQ789", policies[1])
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/TaskQueues/WQ789", policies[2])
        ]:
            yield self.check_policy, method, url, policy

    def test_deprecated_worker(self):
        account_sid = "AC123"
        auth_token = "foobar"
        workspace_sid = "WS456"
        worker_sid = "WK789"
        capability = TaskRouterCapability(account_sid, auth_token, workspace_sid, worker_sid)

        capability.generate_token()

        token = capability.generate_token()
        self.assertNotEqual(None, token)

        decoded = jwt.decode(token, auth_token)
        self.assertNotEqual(None, decoded)

        self.check_decoded(decoded, account_sid, workspace_sid, worker_sid, worker_sid)

        policies = decoded['policies']
        self.assertEqual(len(policies), 5)

        # should expect 5 policies
        for method, url, policy in [
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Activities", policies[0]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Tasks/**", policies[1]),
            ('GET', "https://event-bridge.twilio.com/v1/wschannels/AC123/WK789", policies[2]),
            ('POST', "https://event-bridge.twilio.com/v1/wschannels/AC123/WK789", policies[3]),
            ('GET', "https://taskrouter.twilio.com/v1/Workspaces/WS456/Workers/WK789", policies[4])
        ]:
            yield self.check_policy, method, url, policy

        # check deprecated warnings
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            capability.allow_worker_fetch_attributes()
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "deprecated" in str(w[-1].message)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            capability.allow_worker_activity_updates()
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "deprecated" in str(w[-1].message)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            capability.allow_task_reservation_updates()
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "deprecated" in str(w[-1].message)

if __name__ == "__main__":
    unittest.main()



from mock import patch, Mock
from twilio.rest.resources.ip_messaging import Messages, Message
from tests.tools import create_mock_json

BASE_URI = "https://ip-messaging.twilio.com/v1/Services/ISxxx/Channels/CHxxx"
ACCOUNT_SID = "ACaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
AUTH = (ACCOUNT_SID, "token")
MESSAGE_SID = "MSaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"

list_resource = Messages(BASE_URI, AUTH)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_create_message(mock):
    resp = create_mock_json("tests/resources/ip_messaging/message_instance.json")
    resp.status_code = 201
    mock.return_value = resp

    uri = "%s/Messages" % (BASE_URI)
    list_resource.create('TestBody')
    exp_params = {
        'Body': "TestBody"
    }

    mock.assert_called_with("POST", uri, data=exp_params, auth=AUTH,
                            use_json_extension=False)


@patch("twilio.rest.resources.base.make_twilio_request")
def test_get(mock):
    resp = create_mock_json("tests/resources/ip_messaging/message_instance.json")
    mock.return_value = resp

    uri = "%s/Messages/%s" % (BASE_URI, MESSAGE_SID)
    list_resource.get(MESSAGE_SID)

    mock.assert_called_with("GET", uri, auth=AUTH,
                            use_json_extension=False)


@patch("twilio.rest.resources.base.Resource.request")
def test_delete(req):
    """ Deleting a call should work """
    resp = Mock()
    resp.content = ""
    resp.status_code = 204
    req.return_value = resp, {}

    app = Message(list_resource, "MS123")
    app.delete()
    uri = "%s/Messages/MS123" % (BASE_URI)
    req.assert_called_with("DELETE", uri)



from .members import Members
from .messages import Messages
from twilio.rest.resources import NextGenInstanceResource, NextGenListResource


class Channel(NextGenInstanceResource):

    subresources = [
        Members,
        Messages
    ]

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def delete(self):
        """
        Delete this channel
        """
        return self.delete_instance()


class Channels(NextGenListResource):

    name = "Channels"
    instance = Channel

    def list(self, **kwargs):
        """
        Returns a page of :class:`Channel` resources as a list.
        For paging information see :class:`ListResource`.

        **NOTE**: Due to the potentially voluminous amount of data in an
        alert, the full HTTP request and response data is only returned
        in the Channel instance resource representation.
        """
        return self.get_instances(kwargs)

    def create(self, **kwargs):
        """
        Create a channel.

        :param str friendly_name: The friendly name of the channel.
        :param str attributes: An attribute string with arbitrary

        :return: A :class:`Channel` object
        """
        return self.create_instance(kwargs)

    def delete(self, sid):
        """
        Delete a given Channel
        """
        return self.delete_instance(sid)



from .notifications import Notifications
from .recordings import Recordings
from twilio.rest.resources.call_feedback import (
    CallFeedbackFactory,
    CallFeedbackSummary,
)
from .util import normalize_dates, parse_date, transform_params
from . import InstanceResource, ListResource


class Call(InstanceResource):
    """ A call resource """

    BUSY = "busy"
    CANCELED = "canceled"
    COMPLETED = "completed"
    FAILED = "failed"
    IN_PROGRESS = "in-progress"
    NO_ANSWER = "no-answer"
    QUEUED = "queued"
    RINGING = "ringing"

    subresources = [
        Notifications,
        Recordings,
        CallFeedbackFactory,
    ]

    def hangup(self):
        """ If this call is currenlty active, hang up the call.
        If this call is scheduled to be made, remove the call
        from the queue
        """
        a = self.parent.hangup(self.name)
        self.load(a.__dict__)

    def cancel(self):
        """ If the called is queued or rining, cancel the calls.
        Will not affect in progress calls
        """
        a = self.parent.cancel(self.name)
        self.load(a.__dict__)

    def route(self, **kwargs):
        """Route the specified :class:`Call` to another url.

        :param url: A valid URL that returns TwiML.
        :param method: HTTP method Twilio uses when requesting the above URL.
        """
        a = self.parent.route(self.name, **kwargs)
        self.load(a.__dict__)

    def delete(self):
        """Delete the specified :class:`Call` record from Twilio."""
        return self.parent.delete(self.name)


class Calls(ListResource):
    """ A list of Call resources """

    name = "Calls"
    instance = Call

    def __init__(self, *args, **kwargs):
        super(Calls, self).__init__(*args, **kwargs)
        self.summary = CallFeedbackSummary(self, *args, **kwargs)

    @normalize_dates
    def list(self, from_=None, ended_after=None,
             ended_before=None, ended=None, started_before=None,
             started_after=None, started=None, **kwargs):
        """
        Returns a page of :class:`Call` resources as a list. For paging
        informtion see :class:`ListResource`

        :param date after: Only list calls started after this datetime
        :param date before: Only list calls started before this datetime
        """
        kwargs["from"] = from_
        kwargs["StartTime<"] = started_before
        kwargs["StartTime>"] = started_after
        kwargs["StartTime"] = parse_date(started)
        kwargs["EndTime<"] = ended_before
        kwargs["EndTime>"] = ended_after
        kwargs["EndTime"] = parse_date(ended)
        return self.get_instances(kwargs)

    @normalize_dates
    def iter(self, from_=None, ended_after=None,
             ended_before=None, ended=None, started_before=None,
             started_after=None, started=None, **kwargs):
        """
        Returns an iterator of :class:`Call` resources.

        :param date after: Only list calls started after this datetime
        :param date before: Only list calls started before this datetime
        """
        kwargs["from"] = from_
        kwargs["StartTime<"] = started_before
        kwargs["StartTime>"] = started_after
        kwargs["StartTime"] = parse_date(started)
        kwargs["EndTime<"] = ended_before
        kwargs["EndTime>"] = ended_after
        kwargs["EndTime"] = parse_date(ended)
        return super(Calls, self).iter(**kwargs)

    def create(self, to, from_, url, status_method=None, status_events=None,
               **kwargs):
        """
        Make a phone call to a number.

        :param str to: The phone number to call
        :param str `from_`: The caller ID (must be a verified Twilio number)
        :param str url: The URL to read TwiML from when the call connects
        :param method: The HTTP method Twilio should use to request the url
        :type method: None (defaults to 'POST'), 'GET', or 'POST'
        :param str fallback_url: A URL that Twilio will request if an error
            occurs requesting or executing the TwiML at url
        :param str fallback_method: The HTTP method that Twilio should use
            to request the fallback_url
        :type fallback_method: None (will make 'POST' request),
                               'GET', or 'POST'
        :param str status_callback: A URL that Twilio will request when the
            call ends to notify your app.
        :param str status_method: The HTTP method Twilio should use when
            requesting the above URL.
        :param list status_events: A list of call progress events Twilio
            should send status callback requests on. One or more of:
            'initiated', 'ringing', 'answered', 'completed'. Defaults to
            ['completed'] if not provided. 'completed' events are sent
            free of charge; see twilio.com for current pricing on others.
        :param str if_machine: Tell Twilio to try and determine if a machine
            (like voicemail) or a human has answered the call.
            See more in our `answering machine documentation
            <http://www.twilio.com/docs/api/rest/making_calls>`_.
        :type if_machine: None, 'Continue', or 'Hangup'
        :param str send_digits: A string of keys to dial after
            connecting to the number.
        :type send_digits: None or any combination of
            (0-9), '#', '*' or 'w' (to insert a half second pause).
        :param int timeout: The integer number of seconds that Twilio should
            allow the phone to ring before assuming there is no answer.
        :param str application_sid: The 34 character sid of the application
            Twilio should use to handle this phone call.
            Should not be used in conjunction with the url parameter.

        :return: A :class:`Call` object
        """
        kwargs["from"] = from_
        kwargs["to"] = to
        kwargs["url"] = url
        kwargs["status_callback_method"] = status_method
        kwargs["status_callback_event"] = status_events
        return self.create_instance(kwargs)

    def update(self, sid, **kwargs):
        return self.update_instance(sid, kwargs)

    def cancel(self, sid):
        """ If this call is queued or ringing, cancel the call.
        Will not affect in-progress calls.

        :param sid: A Call Sid for a specific call
        :returns: Updated :class:`Call` resource
        """
        return self.update(sid, status=Call.CANCELED)

    def hangup(self, sid):
        """ If this call is currently active, hang up the call. If this call is
        scheduled to be made, remove the call from the queue.

        :param sid: A Call Sid for a specific call
        :returns: Updated :class:`Call` resource
        """
        return self.update(sid, status=Call.COMPLETED)

    def route(self, sid, url, method="POST"):
        """Route the specified :class:`Call` to another url.

        :param sid: A Call Sid for a specific call
        :param url: A valid URL that returns TwiML.
        :param method: The HTTP method Twilio uses when requesting the URL.
        :returns: Updated :class:`Call` resource
        """
        return self.update(sid, url=url, method=method)

    def feedback(self, sid, quality_score, issue=None):
        """ Create feedback for the given call.

        :param sid: A Call Sid for a specific call
        :param quality_score: The quality of the call
        :param issue: A list of issues experienced during the call
        :returns: A :class:`CallFeedback` object
        """
        uri = "%s/%s" % (self.uri, sid)
        call_feedback_factory = CallFeedbackFactory(
            uri, self.auth, self.timeout
        )
        return call_feedback_factory.create(
            quality_score=quality_score, issue=issue
        )

    def delete(self, sid):
        """Delete the given Call record from Twilio."""
        return self.delete_instance(sid)
